import os
import sys
import pytest
import tempfile
import shutil
import requests
import socket
import psutil
import time
from functools import partial

import numpy as np
import torch
import multiprocessing as mp
from multiprocessing import Queue

from roicat import visualization
import holoviews as hv
from bokeh.server.server import Server
from bokeh.application import Application
from bokeh.application.handlers.function import FunctionHandler

from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options

hv.extension("bokeh")

def get_tempdir_free_space():
    temp_dir = tempfile.gettempdir()
    usage = shutil.disk_usage(temp_dir)
    return usage.free


def create_mock_input():
    ## Create mock input. Looks dumb. But it's just a test.
    mock_data = np.array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=np.float32)
    mock_idx_images_overlay = torch.tensor([0, 1, 2, 3])
    mock_images_overlay = np.random.rand(4, 2, 2)
    return mock_data, mock_idx_images_overlay, mock_images_overlay


def get_indices(path_tempFile):
    ## Steal the fn_get_indices function from roicat.visualization for testing purposes
    with open(path_tempFile, "r") as f:
        indices = f.read().split(",")
    indices = [int(i) for i in indices if i != ""] if len(indices) > 0 else None
    return indices


def deploy_bokeh(instance, tmp_path):
    ## Draw test plot and add to Bokeh document
    hv.extension("bokeh")

    ## Create a mock input
    mock_data, mock_idx_images_overlay, mock_images_overlay = create_mock_input()

    ## Create a scatter plot
    _, layout, _ = visualization.select_region_scatterPlot(
        data=mock_data,
        idx_images_overlay=mock_idx_images_overlay,
        images_overlay=mock_images_overlay,
        size_images_overlay=0.01,
        frac_overlap_allowed=0.5,
        path=tmp_path,
        figsize=(1200, 1200),
        alpha_points=1.0,
        size_points=10,
        color_points="b",
    )

    ## Render plot
    hv_layout = hv.render(layout)
    hv_layout.name = "drawing_test"

    ## Add to Bokeh document
    instance.add_root(hv_layout)


def is_port_available(port, host='localhost'):
    """Check if a given port is available."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        try:
            s.bind((host, port))
        except socket.error:
            return False
        return True


def start_server(apps, query):
    ## Start Bokeh server given a test scatter plot
    server = Server(
        apps,
        port=5006,
        address="0.0.0.0",
        allow_websocket_origin=["0.0.0.0:5006", "localhost:5006"],
    )
    server.start()

    ## Put server details into the queue
    query.put(
        {
            "address": server.address,
            "port": server.port,
        }
    )

    ## Setup IO loop
    server.io_loop.start()


def kill_process_on_port(port):
    for proc in psutil.process_iter(attrs=["pid", "name"]): ## Updated to new psutil API: 2024.09.24
        try:
            connections = proc.connections()
            for conn in connections:
                if conn.status == psutil.CONN_LISTEN and conn.laddr.port == port:
                    try:
                        psutil.Process(proc.info["pid"]).terminate()
                        return True
                    except psutil.AccessDenied:
                        return False
        except (psutil.AccessDenied, psutil.ZombieProcess):
            ## Happens when you don't have permission to access the process
            pass
    return False


def check_server():
    try:
        response = requests.get("http://localhost:5006/test_drawing")
        if response.status_code == 200:
            print("Server is up and running!")
            return 1
        else:
            print(f"Server responded with status code: {response.status_code}")
            return 0
    except requests.ConnectionError:
        print("Cannot connect to the server!")
        return 0


def test_interactive_drawing():
    print("Interactive GUI Drawing Test is running. Please wait...")
    ## Parameter setting
    port = 5006
    server_iter = 0
    max_server_iter = 3
    max_webdriver_iter = 3
    path_tempdir = tempfile.gettempdir()
    path_tempfile = os.path.join(path_tempdir, "indices.csv")

    ## Make sure the tempdir has enough space
    print("Check if tempdir has enough space...")
    free_space = get_tempdir_free_space()
    print(f"Free space in tempdir: {free_space / (1024**3):.2f} GB")

    ## Start iteration. Iteration will stop if indices.csv is created.
    while server_iter <= max_server_iter:
        ## First, check if the port for Bokeh server is available
        if is_port_available(port):
            print(f"Port {port} is available!")
        else:
            print(f"Port {port} is not available!")
            print(f"Kill process using port {port}...")
            kill_process_on_port(port)

        ## Bokeh server deployment at http://localhost:5006/test_drawing
        partial_deploy_bokeh = partial(deploy_bokeh, tmp_path=path_tempfile)
        apps = {"/test_drawing": Application(FunctionHandler(partial_deploy_bokeh))}

        print("Deploy Bokeh server to localhost:5006/test_drawing...")
        ## Let it run in the background so that the test can continue
        query = Queue()
        server_process = mp.Process(target=start_server, args=(apps, query))
        server_process.start()

        ## Wait for the server to start
        time.sleep(5)

        ## Get server info
        server_query = query.get()
        print(f"Server address: {server_query['address']}")
        print(f"Server port: {server_query['port']}")

        ## Prevent permanant hang: If bug occurs, kill the server
        try:
            ## Check if the server is up and running
            print("Check if Bokeh server is up and running...")
            server_status = check_server()
            if not server_status:
                server_process.terminate()
                server_process.join()
                raise Exception("Server is not up and running!")

            print("Setup chrome webdriver...")
            service = Service()
            chrome_options = Options()
            chrome_options.add_argument("--window-size=1280,1280")

            ## For local testing, comment out these options to visualize actions in bokeh / selenium server.
            chrome_options.add_argument("--headless")
            chrome_options.add_argument("--disable-gpu")

            ## if you are on latest version say selenium v4.6.0 or higher, you don't have to use third party library such as WebDriverManager
            print("Driver parameter sanity check...")
            driver = webdriver.Chrome(service=service, options=chrome_options)
            capabilities = driver.capabilities
            print("Browser Name: {}".format(capabilities.get("browserName")))
            print("Browser Version: {}".format(capabilities.get("browserVersion")))
            print("Platform Name: {}".format(capabilities.get("platformName")))
            print(
                "Chrome Driver Version: {}".format(
                    capabilities.get("chrome").get("chromedriverVersion")
                )
            )

            print("Get to the Bokeh server...")
            driver.get("http://localhost:5006/test_drawing")
            wait = WebDriverWait(driver, 10)
            print("Found the Bokeh server, locate drawing Bokeh element...")
            try:
                element = wait.until(EC.presence_of_element_located((By.XPATH, "//*")))
                print("Found Bokeh drawing element!")
            except Exception as e:
                print(f"Failed to locate element: {str(e)}")

            ## Create movement set
            size = element.size
            width, height = size["width"], size["height"]

            print("element size: {}".format(size))
            print("element tagname: {}".format(element.tag_name))
            print("element text: {}".format(element.text))
            print("element location: {}".format(element.location))
            print("element displayed: {}".format(element.is_displayed()))
            print("element enabled: {}".format(element.is_enabled()))
            print("element selected: {}".format(element.is_selected()))

            ## iterate over actions until indices.csv is created
            webdriver_iter = 0
            while webdriver_iter <= max_webdriver_iter:
                print(f"Action iteration {webdriver_iter} starts...")
                ## Move to the center of the element
                print("Start mouse movement...")
                actions = ActionChains(driver)

                ## Surprisingly, this pause seems to be crucial.
                actions.pause(30 * (webdriver_iter + 1))

                ## Move to the center of the element
                actions.move_to_element(element)
                actions.click_and_hold()

                ## Draw!
                actions.move_by_offset(
                    int(width / 2), int(0)
                )  ## Move from center to midpoint of right edge
                actions.move_by_offset(
                    int(0), int(-height / 2)*0.9
                )  ## Move from midpoint of right edge to top right corner
                actions.move_by_offset(
                    int(-width / 2), int(0)
                )  ## Move from top right corner to midpoint of top edge
                actions.release()
                actions.perform()

                print("Mouse movement done!")

                ## Wait for the server to save indices.csv
                time.sleep(5)

                ## Test if indices.csv is created before we kill the server
                if os.path.exists(path_tempfile):
                    print("indices.csv is created! Detach Selenium from Bokeh server...")
                    driver.quit()
                    break
                else:
                    print("indices.csv is not created! Repeat the interaction...")
                    webdriver_iter += 1
        except Exception as e:
            print(f"Exception occured: {e}, Kill the Bokeh server...")
            server_process.terminate()
            server_process.join()

        ## Kill the process to prevent potential race condition
        print("Kill the Bokeh server...")
        server_process.terminate()
        server_process.join()

        ## Wait for the process to terminate
        time.sleep(5)

        ## If indices.csv is created, break the loop
        if os.path.exists(path_tempfile):
            print("indices.csv is created!")
            break
        else:
            print("indices.csv is not created!")
            server_iter += 1

    print("Test if indices.csv has correct permission...")
    if not os.access(path_tempfile, os.R_OK):
        print("indices.csv is not readable!")
        server_process.terminate()
        server_process.join()
        raise Exception("indices.csv is not readable!")

    print("Test if indices are correctly saved...")
    indices = get_indices(path_tempfile)

    if indices is None:
        print("indices.csv is created, but no indices are saved!")
        server_process.terminate()
        server_process.join()
        raise Exception("indices.csv is created, but no indices are saved!")

    ## Check if the indices are correct
    assert indices == [3]
    print("Test is done.")



# Imports
import argparse
import sys
import os
import numpy as np
import torch
import torchvision
from PIL import Image
import matplotlib.pyplot as plt
import json
import pandas as pd
import argparse
import scipy.sparse
from pathlib import Path

from roicat import data_importing
from roicat import ROInet
from roicat.model_training import augmentation
from roicat.model_training import simclr_training_helpers as sth
from roicat.model_training import model

path_script = sys.argv[0]

# Parse arguments for data directory, parameters, and save directory
parser = argparse.ArgumentParser(
    prog='ROICaT SimCLR Training',
    description='This script runs the basic fit pipeline for a self-supervised ROI model using a json file containing the parameters.',
)
parser.add_argument(
    '--directory_data',
    '-d',
    required=False,
    metavar='',
    type=str,
    default=None,
    help='Path to raw ROI data to be used to train the model.',
)
parser.add_argument(
    '--path_params',
    '-p',
    required=False,
    metavar='',
    type=str,
    default=None,
    help='Path to json file containing parameters.',
)
parser.add_argument(
    '--directory_save',
    '-s',
    required=False,
    metavar='',
    type=str,
    default=None,
    help="Directory into which final model and evaluations should be saved.",
)
parser.add_argument(
    '--test_option',
    '-t',
    required=False,
    metavar='',
    type=bool,
    default=False,
    help="Whether to run the script in test mode (True) or not (False). Reduces number of training examples to 3000.",
)
args = parser.parse_args()
directory_data = args.directory_data
filepath_params = args.path_params
directory_save = args.directory_save
test_option = args.test_option

# directory_data = '/media/rich/bigSSD/analysis_data/ROICaT/ROIs_for_training/'
# filepath_params = 'roicat/model_training/simclr_params_base.json'
# directory_save = '/media/rich/bigSSD/analysis_data/ROICaT/training/'
# test_option = True

filepath_logger = str(Path(directory_save) / 'logger.txt')
filepath_losses = str(Path(directory_save) / 'losses.npy')



# Load paths from directory_data and parameters from JSON
list_filepaths_data = [Path(os.path.join(directory_data, filename)) for filename in os.listdir(directory_data)]
with open(filepath_params) as f:
    dict_params = json.load(f)
## If directory_save is specified from the arg parser, overwrite the save directory in the params file
if directory_save is not None:
    dict_params['model']['filepath_model_noPCA'] = str(Path(directory_save) / (dict_params['model']['torchvision_model'] + '_' + 'trainingBest_noPCA.onnx'))

# Load data from dir_data into Data object
ROI_sparse_all = [scipy.sparse.load_npz(filepath_ROI_images) for filepath_ROI_images in list_filepaths_data if filepath_ROI_images.suffix == '.npz']
ROI_images = [sf_sparse.toarray().reshape(sf_sparse.shape[0], 36,36).astype(np.float32) for sf_sparse in ROI_sparse_all]
data = data_importing.Data_roicat();

if test_option:
    print('Running in test mode. Reducing number of training examples.')
    ROI_images = [ROI_images[0][:10000]]
    n_epochs = 2
else:
    n_epochs = dict_params['trainer']['n_epochs']

data.set_ROI_images(
    ROI_images=ROI_images,
    um_per_pixel=dict_params['data']['um_per_pixel'],
)

# Create dataset / dataloader for SimCLR training
ROI_images_rs = ROInet.Resizer_ROI_images(
    ROI_images=np.concatenate(data.ROI_images, axis=0),
    um_per_pixel=dict_params['data']['um_per_pixel'],
    nan_to_num=dict_params['data']['nan_to_num'],
    nan_to_num_val=dict_params['data']['nan_to_num_val'],
    verbose=dict_params['data']['verbose']
).ROI_images_rs
dataloader_generator = ROInet.Dataloader_ROInet(
    ROI_images_rs,
    batchSize_dataloader=dict_params['dataloader']['batchSize_dataloader'],
    pinMemory_dataloader=dict_params['dataloader']['pinMemory_dataloader'],
    numWorkers_dataloader=dict_params['dataloader']['numWorkers_dataloader'],
    persistentWorkers_dataloader=dict_params['dataloader']['persistentWorkers_dataloader'],
    prefetchFactor_dataloader=dict_params['dataloader']['prefetchFactor_dataloader'],
    transforms=torch.nn.Sequential(
        *[augmentation.__dict__[key](**params) for key,params in dict_params['dataloader']['transforms_invariant'].items()]
    ), # Converting dictionary of transforms to torch.nn.Sequential object
    n_transforms=2,
    img_size_out=tuple(dict_params['dataloader']['img_size_out']),
    jit_script_transforms=dict_params['dataloader']['jit_script_transforms'],
    shuffle_dataloader=dict_params['dataloader']['shuffle_dataloader'],
    drop_last_dataloader=dict_params['dataloader']['drop_last_dataloader'],
    verbose=dict_params['dataloader']['verbose'],
)
dataloader = dataloader_generator.dataloader

# Create Model
model_container = model.Simclr_Model(
    filepath_model_save=dict_params['model']['filepath_model_noPCA'], # Set filepath to/from which to save/load model
    base_model=torchvision.models.__dict__[dict_params['model']['torchvision_model']](weights='DEFAULT'),
    head_pool_method=dict_params['model']['head_pool_method'],
    head_pool_method_kwargs=dict_params['model']['head_pool_method_kwargs'],
    pre_head_fc_sizes=dict_params['model']['pre_head_fc_sizes'],
    post_head_fc_sizes=dict_params['model']['post_head_fc_sizes'],
    head_nonlinearity=dict_params['model']['head_nonlinearity'],
    head_nonlinearity_kwargs=dict_params['model']['head_nonlinearity_kwargs'],
    block_to_unfreeze=dict_params['model']['block_to_unfreeze'],
    n_block_toInclude=dict_params['model']['n_block_toInclude'],
    image_out_size=list(dataloader_generator.dataset[0][0][0].shape),
    forward_version=dict_params['trainer']['forward_version'],
)

# Specify criterion, optimizer, scheduler, learning rate, etc.
trainer = sth.Simclr_Trainer(
    dataloader=dataloader,
    model_container=model_container,
    n_epochs=n_epochs,
    device_train=dict_params['trainer']['device_train'],
    inner_batch_size=dict_params['trainer']['inner_batch_size'],
    learning_rate=dict_params['trainer']['learning_rate'],
    penalty_orthogonality=dict_params['trainer']['penalty_orthogonality'],
    weight_decay=dict_params['trainer']['weight_decay'],
    gamma=dict_params['trainer']['gamma'],
    temperature=dict_params['trainer']['temperature'],
    l2_alpha=dict_params['trainer']['l2_alpha'],
    path_saveLog=filepath_logger,
    path_saveLoss=filepath_losses
)

# Loop through epochs, batches, etc. if loss becomes NaNs, don't save the network and stop training.
# Otherwise, save the network as an onnx file.
trainer.train()

if test_option:
    pass
    # trainer.test(torch.ones((1, 3, 224, 224)))

    # trainer.train_pca(torch.rand((260, 3, 224, 224)))
    # trainer.test_pca(torch.ones((1, 3, 224, 224)))

    # trainer.train_pca(torch.cat([torch.cat(x[0], dim=0) for x in trainer.dataloader], axis=0), check_pca_layer_valid=True)



# Documentation style guide:
## - Class init docstrings should be in the class definition docstring.
## - Use a style guide similar to Google's Python style guide except that argument definitions should start on a new indented line after the argument name.
## - If there is more than one argument, use multiple lines for the argument definition code.
## - Example parameters should start on a new line ('\n' should be used before the first one), should start with a dash, and the parameter definition should start on a new indented line.
## - All arguments should have type hints, accurately reflecting the expected type of the argument.
## - Special inputs or conditions related to the arguments should be highlighted using bold for emphasis, italic for optional aspects, and code for specific values or code-related inputs.
## - Keep the return variable name in the docstring for clarity.
## - Keep a consistent line length to improve readability of the docstring.
## - Ensure the clarity of argument descriptions through the use of clear sentence structure and punctuation.

"""
OSF.io links to ROInet versions:

* ROInet_tracking:
    * Info: This version does not include occlusions or large
      affine transformations.
    * Link: https://osf.io/x3fd2/download
    * Hash (MD5 hex): 7a5fb8ad94b110037785a46b9463ea94
* ROInet_classification:
    * Info: This version includes occlusions and large affine
      transformations.
    * Link: https://osf.io/c8m3b/download
    * Hash (MD5 hex): 357a8d9b630ec79f3e015d0056a4c2d5
"""


import sys
from pathlib import Path
import json
import os
import hashlib
import PIL
import multiprocessing as mp
from functools import partial
import gc
from typing import List, Tuple, Union, Optional, Dict, Any, Callable

import numpy as np
import torch
import torchvision
from torch.nn import Module
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
import scipy.signal
import warnings

from . import util, helpers, data_importing

class Resizer_ROI_images(util.ROICaT_Module):
    """
    Class for resizing ROIs.
    RH 2023-2024

    Args:
        function_scaleFactor (Callable):
            The function used to convert ``um_per_pixel`` to a scale factor.
            (Default is ``lambda um_per_pixel, size_im: 1.2 * um_per_pixel * (size_im / 36)``)
            Where ``um_per_pixel`` is the number of microns per pixel and
            size_im is the edge length of the image.
        nan_to_num (bool): 
            Whether to replace NaNs with a specific value. (Default is
            ``True``)
        nan_to_num_val (float): 
            The value to replace NaNs with. (Default is *0.0*)
        verbose (bool): 
            If True, print out extra information. (Default is ``False``)
    """
    def __init__(
        self, 
        function_scaleFactor: Callable[[float, int], float]=lambda um_per_pixel, size_im: 1.2 * um_per_pixel * (size_im / 36),
        nan_to_num: bool=True, 
        nan_to_num_val: float=0.0, verbose: bool=True,
        batch_size: int=10000,
    ):
        super().__init__()
        self.nan_to_num = nan_to_num
        self.nan_to_num_val = nan_to_num_val
        self.batch_size = batch_size
        self._verbose = verbose

        ## Store parameter (but not data) args as attributes
        self.params['__init__'] = self._locals_to_params(
            locals_dict=locals(),
            keys=[
                'nan_to_num',
                'nan_to_num_val',
            ],
        )

        self.function_scaleFactor = function_scaleFactor
        
    def _check_ROI_images(self, ROI_images: np.ndarray):
        ### Check if any NaNs
        if np.any(np.isnan(ROI_images)):
            if self.nan_to_num:
                warnings.warn('ROICaT WARNING: NaNs detected. You should consider removing these before passing to the network. Using nan_to_num arguments.')
            else:
                raise ValueError('ROICaT ERROR: NaNs detected. You should consider removing these before passing to the network. Use nan_to_num=True to ignore this error.')
        if np.any(np.isinf(ROI_images)):
            warnings.warn('ROICaT WARNING: Infs detected. You should consider removing these before passing to the network.')
        ## Check if any images in any of the sessions are all zeros
        if np.any(np.all(ROI_images==0, axis=(1,2))):
            warnings.warn('ROICaT WARNING: Image(s) with all zeros detected. These can pass through the network, but may give weird results.')

    def plot_resized_comparison(self, ROI_images_cat: np.ndarray, ROI_images_rs: np.ndarray):
        """
        Plot a comparison of the ROI sizes before and after resizing.

        Args:
            ROI_images_cat (np.ndarray):
                Array of ROIs to resize. Shape should be (nROIs, height,
                width).
            ROI_images_rs (np.ndarray):
                Array of resized ROIs. Shape should be (nROIs, height, width).
        """
        fig, axs = plt.subplots(2,1, figsize=(7,10))
        axs[0].plot(np.mean(ROI_images_cat > 0, axis=(1,2)))
        axs[0].plot(scipy.signal.savgol_filter(np.mean(ROI_images_cat > 0, axis=(1,2)), 501, 3))
        axs[0].set_xlabel('ROI number');
        axs[0].set_ylabel('mean npix');
        axs[0].set_title('ROI sizes raw')

        axs[1].plot(np.mean(ROI_images_rs > 0, axis=(1,2)))
        axs[1].plot(scipy.signal.savgol_filter(np.mean(ROI_images_rs > 0, axis=(1,2)), 501, 3))
        axs[1].set_xlabel('ROI number');
        axs[1].set_ylabel('mean npix');
        axs[1].set_title('ROI sizes resized')

    def resize_ROIs(
        self,
        ROI_images: np.ndarray,  # Array of shape (n_rois, height, width)
        um_per_pixel: float,
    ) -> np.ndarray:
        """
        Resizes the ROI (Region of Interest) images to prepare them for pass
        through network.

        Args:
            ROI_images (np.ndarray): 
                The ROI images to resize. Array of shape *(n_rois, height,
                width)*.
            um_per_pixel (float): 
                The number of microns per pixel. This value is used to rescale
                the ROI images so that they occupy a standard region of the
                image frame.

        Returns:
            (np.ndarray): 
                ROI_images_rs (np.ndarray): 
                    The resized ROI images.
        """
        ## Store parameter (but not data) args as attributes
        self.params['resize_ROIs'] = self._locals_to_params(
            locals_dict=locals(),
            keys=[
                'um_per_pixel',
            ],
        )

        self._check_ROI_images(ROI_images)
        assert isinstance(um_per_pixel, (int, float)), f'um_per_pixel should be an int or float, but is {type(um_per_pixel)}'

        if self.nan_to_num:
            print(f'ROICaT: replacing NaNs with {self.nan_to_num_val}') if self._verbose else None
            ROI_images = np.nan_to_num(ROI_images, nan=self.nan_to_num_val)

        scale_forRS = self.function_scaleFactor(um_per_pixel=float(um_per_pixel), size_im=ROI_images.shape[1])

        print(f'ROICaT: resizing ROIs') if self._verbose else None
        return np.stack([resize_affine(img, scale=scale_forRS, clamp_range=True) for img in tqdm(ROI_images, mininterval=5, disable=not self._verbose)], axis=0)
        ## Faster but slightly different results
        # return np.concatenate(
        #     [resize_images(
        #         batch, 
        #         scale=scale_forRS, 
        #         clamp_range=True,
        #     ) for batch in tqdm(
        #         helpers.make_batches(ROI_images, batch_size=self.batch_size), 
        #         total=np.ceil(len(ROI_images)/self.batch_size), 
        #         mininterval=5, 
        #         unit='images',
        #         unit_scale=self.batch_size,
        #         disable=not self._verbose,
        #     )], axis=0)


class Dataloader_ROInet(util.ROICaT_Module):
    """
    Class for creating a dataloader for the ROInet network.
    JZ, RH 2023
        
    Args:
        ROI_images (np.ndarray):
            Array of ROIs to resize. Shape should be (nROIs, height,
            width).
        pref_plot (bool): 
            If ``True``, plots the sizes of the ROI images before and after
            normalization. (Default is ``False``)
        batchSize_dataloader (int): 
            The batch size to use for the DataLoader. (Default is *8*)
        pinMemory_dataloader (bool): 
            If ``True``, pins the memory of the DataLoader, as per PyTorch's
            best practices. (Default is ``True``)
        numWorkers_dataloader (int): 
            The number of worker processes for data loading. (Default is
            *-1*)
        persistentWorkers_dataloader (bool): 
            If ``True``, uses persistent worker processes. (Default is
            ``True``)
        prefetchFactor_dataloader (int): 
            The prefetch factor for data loading. (Default is *2*)
        transforms (Optional[Callable]): 
            The transforms to use for the DataLoader. If ``None``, the
            function will only scale dynamic range (to 0-1), resize (to
            img_size_out dimensions), and tile channels (to 3) as a minimum
            to pass images through the network. (Default is ``None``)
        n_transforms (int):
            The number of times to apply the transforms to each image. Should
            be 1 for inference and 2 for training. (Default is *1*)
        img_size_out (Tuple[int, int]): 
            The image output dimensions of DataLoader if transforms is
            ``None``. (Default is *(224, 224)*)
        jit_script_transforms (bool): 
            If ``True``, converts the transforms pipeline into a TorchScript
            pipeline, potentially improving calculation speed but can cause
            problems with multiprocessing. (Default is ``False``)
        shuffle (bool):
            If ``True``, shuffles the data. Should be set to ``True`` for
            SimCLR training. (Default is ``False``)
        drop_last (bool):
            If ``True``, drops the last batch if it is not full. Should be
            set to ``True`` for SimCLR training. (Default is ``False``)
        verbose (bool):
            If ``True``, print out extra information. (Default is ``True``)    """
    def __init__(
            self,
            ROI_images: np.ndarray,
            batchSize_dataloader: int = 8,
            pinMemory_dataloader: bool = True,
            numWorkers_dataloader: int = -1,
            persistentWorkers_dataloader: bool = True,
            prefetchFactor_dataloader: int = 2,
            transforms: Optional[Callable] = None,
            n_transforms: int = 1,
            img_size_out: Tuple[int, int] = (224, 224),
            jit_script_transforms: bool = False,
            shuffle_dataloader: bool = False,
            drop_last_dataloader: bool = False,
            verbose: bool = True,
        ):
        super().__init__()

        self._verbose = verbose
        numWorkers_dataloader = mp.cpu_count() if numWorkers_dataloader == -1 else numWorkers_dataloader

        ## Store parameter (but not data) args as attributes
        self.params['__init__'] = self._locals_to_params(
            locals_dict=locals(),
            keys=[
                'batchSize_dataloader',
                'pinMemory_dataloader',
                'numWorkers_dataloader',
                'persistentWorkers_dataloader',
                'prefetchFactor_dataloader',
                'n_transforms',
                'img_size_out',
                'jit_script_transforms',
                'shuffle_dataloader',
                'drop_last_dataloader',
                'verbose',
            ],
        )

        ## Type checking / correction
        if not isinstance(img_size_out, (tuple, list)):
            assert isinstance(img_size_out, int), f'img_size_out should be a tuple or list, but is {type(img_size_out)}'
            img_size_out = (img_size_out, img_size_out)

        transforms = torch.nn.Sequential(
            ScaleDynamicRange(scaler_bounds=(0,1)),
            torchvision.transforms.Resize(
                size=img_size_out,
                interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
                antialias=True,
            ),
            TileChannels(dim=0, n_channels=3),
        ) if transforms is None else transforms

        if jit_script_transforms:
            if numWorkers_dataloader > 0:
                warnings.warn("\n\nWarning: Converting transforms to a jit-based script has been known to cause issues on Windows when numWorkers_dataloader > 0. If self.generate_latents() raises an Exception similar to 'Tried to serialize object __torch__.torch.nn.modules.container.Sequential which does not have a __getstate__ method defined!' consider setting numWorkers_dataloader=0 or jit_script_transforms=False.\n")
            self.transforms = torch.jit.script(transforms)
        else:
            self.transforms = transforms
        
        print(f'Defined image transformations: {transforms}') if self._verbose else None
        self.dataset = dataset_simCLR(
                X=torch.as_tensor(ROI_images, device='cpu', dtype=torch.float32),
                y=torch.as_tensor(torch.zeros(ROI_images.shape[0]), device='cpu', dtype=torch.float32),
                n_transforms=n_transforms,
                transform=self.transforms,
                DEVICE='cpu',
                dtype_X=torch.float32,
            )
        print(f'Defined dataset') if self._verbose else None
        self.dataloader = torch.utils.data.DataLoader(
                self.dataset,
                batch_size=batchSize_dataloader,
                shuffle=shuffle_dataloader,
                drop_last=drop_last_dataloader,
                pin_memory=pinMemory_dataloader,
                num_workers=numWorkers_dataloader,
                persistent_workers=persistentWorkers_dataloader,
                prefetch_factor=prefetchFactor_dataloader,
        )
        print(f'Defined dataloader') if self._verbose else None


class ROInet_embedder(util.ROICaT_Module):
    """
    Class for loading the ROInet model, preparing data for it, and running it.
    RH, JZ 2022
    
    OSF.io links to ROInet versions:

    * ROInet_tracking:
        * Info: This version does not include occlusions or large affine
          transformations.
        * Link: https://osf.io/x3fd2/download
        * Hash (MD5 hex): 7a5fb8ad94b110037785a46b9463ea94
    * ROInet_classification:
        * Info: This version includes occlusions and large affine
          transformations.
        * Link: https://osf.io/c8m3b/download
        * Hash (MD5 hex): 357a8d9b630ec79f3e015d0056a4c2d5
    
    Args:
        dir_networkFiles (str): 
            Directory to find an existing ROInet.zip file or download and
            extract a new one into.
        device (str): 
            Device to use for the model and data. (Default is ``'cpu'``)
        download_method (str): 
            Approach to downloading the network files. Options are: \n
            * ``'check_local_first'``: Check if the network files are already in
              dir_networkFiles, if so, use them.
            * ``'force_download'``: Download an ROInet.zip file from
              download_url.
            * ``'force_local'``: Use an existing local copy of an ROInet.zip
              file, if they don't exist, raise an error. Hash checking is done
              and download_hash must be specified. \n
            (Default is ``'check_local_first'``)
        download_url (str): 
            URL to download the ROInet.zip file from.
            (Default is https://osf.io/x3fd2/download)
        download_hash (dict): 
            MD5 hash of the ROInet.zip file. This can be obtained from
            ROICaT documentation. If you don't have one, use
            download_method='force_download' and determine the hash using
            helpers.hash_file(). (Default is ``None``)
        names_networkFiles (dict): 
            Names of the files in the ROInet.zip file. If uncertain, leave
            as None. The dictionary should have the form: \n
            ``{'params': 'params.json', 'model': 'model.py', 'state_dict':
            'ConvNext_tiny__1_0_unfrozen__simCLR.pth',}`` \n
            Where 'params' is the parameters used to train the network
            (usually a .json file), 'model' is the model definition (usually
            a .py file), and 'state_dict' are the weights of the network
            (usually a .pth file). (Default is ``None``)
        forward_pass_version (str): 
            Version of the forward pass to use. Options are 'latent' (return
            the post-head output latents, use this for tracking), 'head'
            (return the output of the head layers, use this for
            classification), and 'base' (return the output of the base
            model). (Default is ``'latent'``)
        verbose (bool): 
            If True, print out extra information. (Default is ``True``)
    """
    def __init__(
        self,
        dir_networkFiles: str,
        device: str = 'cpu',
        download_method: str = 'check_local_first',
        download_url: str = 'https://osf.io/x3fd2/download',
        download_hash: dict = None,
        names_networkFiles: dict = None,
        forward_pass_version: str = 'latent',
        verbose: bool = True,
    ):
        ## Imports
        super().__init__()

        ## Store parameter (but not data) args as attributes
        self.params['__init__'] = self._locals_to_params(
            locals_dict=locals(),
            keys=[
                'dir_networkFiles',
                'device',
                'download_method',
                'download_url',
                'download_hash',
                'names_networkFiles',
                'forward_pass_version',
                'verbose',
            ],
        )

        self._device = device
        self._verbose = verbose
        self._dir_networkFiles = dir_networkFiles
        self._download_url = download_url

        self._download_path_save = str(Path(self._dir_networkFiles).resolve() / 'ROInet.zip')

        fn_download = partial(
            helpers.download_file,
            path_save=self._download_path_save,
            hash_type='MD5',
            hash_hex=download_hash,
            mkdir=True,
            allow_overwrite=True,
            write_mode='wb',
            verbose=self._verbose,
            chunk_size=1024,
        )

        ## Find or download network files
        if download_method == 'force_download':
            fn_download(url=self._download_url, check_local_first=False, check_hash=False)

        if download_method == 'check_local_first':
            # assert download_hash is not None, "if using download_method='check_local_first' download_hash cannot be None. Either determine the hash of the zip file or use download_method='force_download'."
            fn_download(url=self._download_url, check_local_first=True, check_hash=True)

        if download_method == 'force_local':
            # assert download_hash is not None, "if using download_method='force_local' download_hash cannot be None"
            assert Path(self._download_path_save).exists(), f"if using download_method='force_local' the network files must exist in {self._download_path_save}"
            fn_download(url=None, check_local_first=True, check_hash=True)

        ## Extract network files from zip
        paths_extracted = helpers.extract_zip(
            path_zip=self._download_path_save,
            path_extract=self._dir_networkFiles,
            verbose=self._verbose,
        )

        ## Find network files
        if names_networkFiles is None:
            names_networkFiles = {
                'params': 'params.json',
                'model': 'model.py',
                'state_dict': '.pth',
            }
        paths_networkFiles = {}
        paths_networkFiles['params'] = [p for p in paths_extracted if names_networkFiles['params'] in str(Path(p).name)][0]
        paths_networkFiles['model'] = [p for p in paths_extracted if names_networkFiles['model'] in str(Path(p).name)][0]
        paths_networkFiles['state_dict'] = [p for p in paths_extracted if names_networkFiles['state_dict'] in str(Path(p).name)][0]

        ## Import network files
        sys.path.append(str(Path(paths_networkFiles['model']).parent.resolve()))
        import model
        print(f"Imported model from {paths_networkFiles['model']}") if self._verbose else None

        with open(paths_networkFiles['params']) as f:
            self.params_model = json.load(f)
            print(f"Loaded params_model from {paths_networkFiles['params']}") if self._verbose else None
            self.net = model.make_model(fwd_version=forward_pass_version, **self.params_model)
            print(f"Generated network using params_model") if self._verbose else None

        ## Prep network and load state_dict
        for param in self.net.parameters():
            param.requires_grad = False
        self.net.eval()

        self.net.load_state_dict(torch.load(
            f=paths_networkFiles['state_dict'], 
            map_location=torch.device(self._device),
            weights_only=True,
        ))
        print(f'Loaded state_dict into network from {paths_networkFiles["state_dict"]}') if self._verbose else None

        self.net = self.net.to(self._device)
        print(f'Loaded network onto device {self._device}') if self._verbose else None

    def generate_dataloader(
        self,
        ROI_images: List[np.ndarray],
        um_per_pixel: Union[float, List[float]],
        resize_ROI_images: bool = True,
        nan_to_num: bool = True,
        nan_to_num_val: float = 0.0,
        pref_plot: bool = False,
        batchSize_dataloader: int = 8,
        pinMemory_dataloader: bool = True,
        numWorkers_dataloader: int = -1,
        persistentWorkers_dataloader: bool = True,
        prefetchFactor_dataloader: int = 2,
        transforms: Optional[Callable] = None,
        img_size_out: Tuple[int, int] = (224, 224),
        jit_script_transforms: bool = False,
    ):
        """
        Generates a PyTorch DataLoader for a list of Region of Interest (ROI)
        images. Performs preprocessing such as rescaling, normalization, and
        resizing.

        Args:
            ROI_images (List[np.ndarray]): 
                The ROI images to use for the dataloader. List of arrays, each
                array corresponds to a session and is of shape *(n_rois, height,
                width)*.
            um_per_pixel (Union[float, List[float]]):
                The conversion factor from pixels to microns. This is used to scale
                the ROI_images to a common size. Should either be a float or a list
                of floats, one for each session.
            resize_ROI_images (bool):
                If ``True``, resizes the ROI images to a common size. (Default is
                ``True``)
            nan_to_num (bool): 
                Whether to replace NaNs with a specific value. (Default is
                ``True``)
            nan_to_num_val (float): 
                The value to replace NaNs with. (Default is *0.0*)
            pref_plot (bool): 
                If ``True``, plots the sizes of the ROI images before and after
                normalization. (Default is ``False``)
            batchSize_dataloader (int): 
                The batch size to use for the DataLoader. (Default is *8*)
            pinMemory_dataloader (bool): 
                If ``True``, pins the memory of the DataLoader, as per PyTorch's
                best practices. (Default is ``True``)
            numWorkers_dataloader (int): 
                The number of worker processes for data loading. (Default is
                *-1*)
            persistentWorkers_dataloader (bool): 
                If ``True``, uses persistent worker processes. (Default is
                ``True``)
            prefetchFactor_dataloader (int): 
                The prefetch factor for data loading. (Default is *2*)
            transforms (Optional[Callable]): 
                The transforms to use for the DataLoader. If ``None``, the
                function will only scale dynamic range (to 0-1), resize (to
                img_size_out dimensions), and tile channels (to 3) as a minimum
                to pass images through the network. (Default is ``None``)
            img_size_out (Tuple[int, int]): 
                The image output dimensions of DataLoader if transforms is
                ``None``. (Default is *(224, 224)*)
            jit_script_transforms (bool): 
                If ``True``, converts the transforms pipeline into a TorchScript
                pipeline, potentially improving calculation speed but can cause
                problems with multiprocessing. (Default is ``False``)

        Returns:
            (np.ndarray): 
                ROI_images (np.ndarray): 
                    The ROI images after normalization and resizing. Shape is
                    *(n_sessions, n_rois, n_channels, height, width)*.

        Example:
            .. highlight:: python
            .. code-block:: python

                dataloader = generate_dataloader(ROI_images)
        """
        um_per_pixel = data_importing.Data_roicat._fix_um_per_pixel(um_per_pixel=um_per_pixel, n_sessions=len(ROI_images))
        ROI_images = data_importing.Data_roicat._fix_ROI_images(ROI_images=ROI_images)

        ## Store parameter (but not data) args as attributes
        self.params['generate_dataloader'] = self._locals_to_params(
            locals_dict=locals(),
            keys=[
                'um_per_pixel',
                'nan_to_num',
                'nan_to_num_val',
                'pref_plot',
                'batchSize_dataloader',
                'pinMemory_dataloader',
                'numWorkers_dataloader',
                'persistentWorkers_dataloader',
                'prefetchFactor_dataloader',
                'img_size_out',
                'jit_script_transforms',
            ],
        )    

        if resize_ROI_images:
            print(f'Starting Image Resizer') if self._verbose else None
            roi_resizer = Resizer_ROI_images(
                nan_to_num=nan_to_num,
                nan_to_num_val=nan_to_num_val,
                verbose=False,
            )
            self.ROI_images_rs = np.concatenate([
                roi_resizer.resize_ROIs(
                    ROI_images=ROI_images[ii], 
                    um_per_pixel=um_per_pixel[ii],
                ) for ii in range(len(ROI_images))
            ], axis=0)

            roi_resizer.plot_resized_comparison(
                ROI_images_cat=np.concatenate(ROI_images, axis=0),
                ROI_images_rs=self.ROI_images_rs,
            ) if pref_plot else None
        else:
            self.ROI_images_rs = np.concatenate(ROI_images, axis=0)

        print(f'Creating dataloader') if self._verbose else None
        dataloader_generator = Dataloader_ROInet(
            ROI_images=self.ROI_images_rs,
            batchSize_dataloader=batchSize_dataloader,
            pinMemory_dataloader=pinMemory_dataloader,
            numWorkers_dataloader=numWorkers_dataloader,
            persistentWorkers_dataloader=persistentWorkers_dataloader,
            prefetchFactor_dataloader=prefetchFactor_dataloader,
            transforms=transforms,
            n_transforms=1,
            img_size_out=img_size_out,
            jit_script_transforms=jit_script_transforms,
            shuffle_dataloader=False,
            drop_last_dataloader=False,
            verbose=self._verbose,
        )

        self.transforms = dataloader_generator.transforms
        self.dataset = dataloader_generator.dataset
        self.dataloader = dataloader_generator.dataloader
        return self.ROI_images_rs

    def generate_latents(self) -> torch.Tensor:
        """
        Passes the data in the dataloader through the network and generates latents.

        Returns:
            (torch.Tensor): 
                latents (torch.Tensor): 
                    Latents for each ROI (Region of Interest).
        """
        if hasattr(self, 'dataloader') == False:
            raise Exception('dataloader not defined. Call generate_dataloader() first.')

        print(f'starting: running data through network')
        self.latents = torch.cat([self.net(data[0][0].to(self._device)).detach() for data in tqdm(self.dataloader, mininterval=5)], dim=0).cpu()
        print(f'completed: running data through network')

        gc.collect()
        torch.cuda.empty_cache()
        gc.collect()
        torch.cuda.empty_cache()
        
        return self.latents


class ROInet_embedder_original(util.ROICaT_Module):
    """
    Class for loading the ROInet model, preparing data for it, and running it.
    RH, JZ 2022
    
    OSF.io links to ROInet versions:

    * ROInet_tracking:
        * Info: This version does not include occlusions or large affine
          transformations.
        * Link: https://osf.io/x3fd2/download
        * Hash (MD5 hex): 7a5fb8ad94b110037785a46b9463ea94
    * ROInet_classification:
        * Info: This version includes occlusions and large affine
          transformations.
        * Link: https://osf.io/c8m3b/download
        * Hash (MD5 hex): 357a8d9b630ec79f3e015d0056a4c2d5
    
    Args:
        dir_networkFiles (str): 
            Directory to find an existing ROInet.zip file or download and
            extract a new one into.
        device (str): 
            Device to use for the model and data. (Default is ``'cpu'``)
        download_method (str): 
            Approach to downloading the network files. Options are: \n
            * ``'check_local_first'``: Check if the network files are already in
              dir_networkFiles, if so, use them.
            * ``'force_download'``: Download an ROInet.zip file from
              download_url.
            * ``'force_local'``: Use an existing local copy of an ROInet.zip
              file, if they don't exist, raise an error. Hash checking is done
              and download_hash must be specified. \n
            (Default is ``'check_local_first'``)
        download_url (str): 
            URL to download the ROInet.zip file from.
            (Default is https://osf.io/x3fd2/download)
        download_hash (dict): 
            MD5 hash of the ROInet.zip file. This can be obtained from
            ROICaT documentation. If you don't have one, use
            download_method='force_download' and determine the hash using
            helpers.hash_file(). (Default is ``None``)
        names_networkFiles (dict): 
            Names of the files in the ROInet.zip file. If uncertain, leave
            as None. The dictionary should have the form: \n
            ``{'params': 'params.json', 'model': 'model.py', 'state_dict':
            'ConvNext_tiny__1_0_unfrozen__simCLR.pth',}`` \n
            Where 'params' is the parameters used to train the network
            (usually a .json file), 'model' is the model definition (usually
            a .py file), and 'state_dict' are the weights of the network
            (usually a .pth file). (Default is ``None``)
        forward_pass_version (str): 
            Version of the forward pass to use. Options are 'latent' (return
            the post-head output latents, use this for tracking), 'head'
            (return the output of the head layers, use this for
            classification), and 'base' (return the output of the base
            model). (Default is ``'latent'``)
        verbose (bool): 
            If True, print out extra information. (Default is ``True``)
    """
    def __init__(
        self,
        dir_networkFiles: str,
        device: str = 'cpu',
        download_method: str = 'check_local_first',
        download_url: str = 'https://osf.io/x3fd2/download',
        download_hash: dict = None,
        names_networkFiles: dict = None,
        forward_pass_version: str = 'latent',
        verbose: bool = True,
    ):
        ## Imports
        super().__init__()

        self._device = device
        self._verbose = verbose


        self._dir_networkFiles = dir_networkFiles
        self._download_url = download_url

        self._download_path_save = str(Path(self._dir_networkFiles).resolve() / 'ROInet.zip')

        fn_download = partial(
            helpers.download_file,
            path_save=self._download_path_save,
            hash_type='MD5',
            hash_hex=download_hash,
            mkdir=True,
            allow_overwrite=True,
            write_mode='wb',
            verbose=self._verbose,
            chunk_size=1024,
        )

        ## Find or download network files
        if download_method == 'force_download':
            fn_download(url=self._download_url, check_local_first=False, check_hash=False)

        if download_method == 'check_local_first':
            # assert download_hash is not None, "if using download_method='check_local_first' download_hash cannot be None. Either determine the hash of the zip file or use download_method='force_download'."
            fn_download(url=self._download_url, check_local_first=True, check_hash=True)

        if download_method == 'force_local':
            # assert download_hash is not None, "if using download_method='force_local' download_hash cannot be None"
            assert Path(self._download_path_save).exists(), f"if using download_method='force_local' the network files must exist in {self._download_path_save}"
            fn_download(url=None, check_local_first=True, check_hash=True)

        ## Extract network files from zip
        paths_extracted = helpers.extract_zip(
            path_zip=self._download_path_save,
            path_extract=self._dir_networkFiles,
            verbose=self._verbose,
        )

        ## Find network files
        if names_networkFiles is None:
            names_networkFiles = {
                'params': 'params.json',
                'model': 'model.py',
                'state_dict': '.pth',
            }
        paths_networkFiles = {}
        paths_networkFiles['params'] = [p for p in paths_extracted if names_networkFiles['params'] in str(Path(p).name)][0]
        paths_networkFiles['model'] = [p for p in paths_extracted if names_networkFiles['model'] in str(Path(p).name)][0]
        paths_networkFiles['state_dict'] = [p for p in paths_extracted if names_networkFiles['state_dict'] in str(Path(p).name)][0]

        ## Import network files
        sys.path.append(str(Path(paths_networkFiles['model']).parent.resolve()))
        import model
        print(f"Imported model from {paths_networkFiles['model']}") if self._verbose else None

        with open(paths_networkFiles['params']) as f:
            self.params_model = json.load(f)
            print(f"Loaded params_model from {paths_networkFiles['params']}") if self._verbose else None
            self.net = model.make_model(fwd_version=forward_pass_version, **self.params_model)
            print(f"Generated network using params_model") if self._verbose else None

        ## Prep network and load state_dict
        for param in self.net.parameters():
            param.requires_grad = False
        self.net.eval()

        self.net.load_state_dict(torch.load(paths_networkFiles['state_dict'], map_location=torch.device(self._device)))
        print(f'Loaded state_dict into network from {paths_networkFiles["state_dict"]}') if self._verbose else None

        self.net = self.net.to(self._device)
        print(f'Loaded network onto device {self._device}') if self._verbose else None

    def generate_dataloader(
        self,
        ROI_images: List[np.ndarray],
        um_per_pixel: float = 1.0,
        nan_to_num: bool = True,
        nan_to_num_val: float = 0.0,
        pref_plot: bool = False,
        batchSize_dataloader: int = 8,
        pinMemory_dataloader: bool = True,
        numWorkers_dataloader: int = -1,
        persistentWorkers_dataloader: bool = True,
        prefetchFactor_dataloader: int = 2,
        transforms: Optional[Callable] = None,
        img_size_out: Tuple[int, int] = (224, 224),
        jit_script_transforms: bool = False,
    ):
        """
        Generates a PyTorch DataLoader for a list of Region of Interest (ROI)
        images. Performs preprocessing such as rescaling, normalization, and
        resizing.

        Args:
            ROI_images (List[np.ndarray]): 
                The ROI images to use for the dataloader. List of arrays, each
                array corresponds to a session and is of shape *(n_rois, height,
                width)*.
            um_per_pixel (float): 
                The number of microns per pixel. Used to rescale the ROI images
                to the same size as the network input. (Default is *1.0*)
            nan_to_num (bool): 
                Whether to replace NaNs with a specific value. (Default is
                ``True``)
            nan_to_num_val (float): 
                The value to replace NaNs with. (Default is *0.0*)
            pref_plot (bool): 
                If ``True``, plots the sizes of the ROI images before and after
                normalization. (Default is ``False``)
            batchSize_dataloader (int): 
                The batch size to use for the DataLoader. (Default is *8*)
            pinMemory_dataloader (bool): 
                If ``True``, pins the memory of the DataLoader, as per PyTorch's
                best practices. (Default is ``True``)
            numWorkers_dataloader (int): 
                The number of worker processes for data loading. (Default is
                *-1*)
            persistentWorkers_dataloader (bool): 
                If ``True``, uses persistent worker processes. (Default is
                ``True``)
            prefetchFactor_dataloader (int): 
                The prefetch factor for data loading. (Default is *2*)
            transforms (Optional[Callable]): 
                The transforms to use for the DataLoader. If ``None``, the
                function will only scale dynamic range (to 0-1), resize (to
                img_size_out dimensions), and tile channels (to 3) as a minimum
                to pass images through the network. (Default is ``None``)
            img_size_out (Tuple[int, int]): 
                The image output dimensions of DataLoader if transforms is
                ``None``. (Default is *(224, 224)*)
            jit_script_transforms (bool): 
                If ``True``, converts the transforms pipeline into a TorchScript
                pipeline, potentially improving calculation speed but can cause
                problems with multiprocessing. (Default is ``False``)

        Returns:
            (np.ndarray): 
                ROI_images (np.ndarray): 
                    The ROI images after normalization and resizing. Shape is
                    *(n_sessions, n_rois, n_channels, height, width)*.

        Example:
            .. highlight:: python
            .. code-block:: python

                dataloader = generate_dataloader(ROI_images)
        """
        ## Remove NaNs
        ### Check if any NaNs
        if np.any([np.any(np.isnan(roi)) for roi in ROI_images]):
            warnings.warn('ROICaT WARNING: NaNs detected. You should consider removing remove these before passing to the network. Using nan_to_num arguments.')
        if np.any([np.any(np.isinf(roi)) for roi in ROI_images]):
            warnings.warn('ROICaT WARNING: Infs detected. You should consider removing these before passing to the network.')
        ## Check if any images in any of the sessions are all zeros
        if np.any([np.any(np.all(rois==0, axis=(1,2))) for rois in ROI_images]):
            warnings.warn('ROICaT WARNING: Image(s) with all zeros detected. These can pass through the network, but may give weird results.')
        if nan_to_num:
            ROI_images = [np.nan_to_num(rois, nan=nan_to_num_val) for rois in ROI_images]

        if numWorkers_dataloader == -1:
            numWorkers_dataloader = mp.cpu_count()

        print('Starting: resizing ROIs') if self._verbose else None
        
        sf_rs = [self.resize_ROIs(rois, um_per_pixel) for rois in ROI_images]
        
        ROI_images_cat = np.concatenate(ROI_images, axis=0)
        ROI_images_rs = np.concatenate(sf_rs, axis=0)

        print('Completed: resizing ROIs') if self._verbose else None

        if pref_plot:
            fig, axs = plt.subplots(2,1, figsize=(7,10))
            axs[0].plot(np.mean(ROI_images_cat > 0, axis=(1,2)))
            axs[0].plot(scipy.signal.savgol_filter(np.mean(ROI_images_cat > 0, axis=(1,2)), 501, 3))
            axs[0].set_xlabel('ROI number');
            axs[0].set_ylabel('mean npix');
            axs[0].set_title('ROI sizes raw')

            axs[1].plot(np.mean(ROI_images_rs > 0, axis=(1,2)))
            axs[1].plot(scipy.signal.savgol_filter(np.mean(ROI_images_rs > 0, axis=(1,2)), 501, 3))
            axs[1].set_xlabel('ROI number');
            axs[1].set_ylabel('mean npix');
            axs[1].set_title('ROI sizes resized')

        if transforms is None:
            transforms = torch.nn.Sequential(
                ScaleDynamicRange(scaler_bounds=(0,1)),
                torchvision.transforms.Resize(
                    size=img_size_out,
                    interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
                    antialias=True,
                ),
                TileChannels(dim=0, n_channels=3),
            )

        if jit_script_transforms:
            if numWorkers_dataloader > 0:
                warnings.warn("\n\nWarning: Converting transforms to a jit-based script has been known to cause issues on Windows when numWorkers_dataloader > 0. If self.generate_latents() raises an Exception similar to 'Tried to serialize object __torch__.torch.nn.modules.container.Sequential which does not have a __getstate__ method defined!' consider setting numWorkers_dataloader=0 or jit_script_transforms=False.\n")

            self.transforms = torch.jit.script(transforms)
        else:
            self.transforms = transforms
        
        print(f'Defined image transformations: {transforms}') if self._verbose else None

        self.dataset = dataset_simCLR(
                X=torch.as_tensor(ROI_images_rs, device='cpu', dtype=torch.float32),
                y=torch.as_tensor(torch.zeros(ROI_images_rs.shape[0]), device='cpu', dtype=torch.float32),
                n_transforms=1,
                transform=self.transforms,
                DEVICE='cpu',
                dtype_X=torch.float32,
            )
        print(f'Defined dataset') if self._verbose else None

        self.dataloader = torch.utils.data.DataLoader(
                self.dataset,
                batch_size=batchSize_dataloader,
                shuffle=False,
                drop_last=False,
                pin_memory=pinMemory_dataloader,
                num_workers=numWorkers_dataloader,
                persistent_workers=persistentWorkers_dataloader,
                prefetch_factor=prefetchFactor_dataloader,
        )

        print(f'Defined dataloader') if self._verbose else None

        self.ROI_images_rs = ROI_images_rs
        return ROI_images_rs

    @classmethod
    def resize_ROIs(
        cls,
        ROI_images: np.ndarray,  # Array of shape (n_rois, height, width)
        um_per_pixel: float,
    ) -> np.ndarray:
        """
        Resizes the ROI (Region of Interest) images to prepare them for pass
        through network.

        Args:
            ROI_images (np.ndarray): 
                The ROI images to resize. Array of shape *(n_rois, height,
                width)*.
            um_per_pixel (float): 
                The number of microns per pixel. This value is used to rescale
                the ROI images so that they occupy a standard region of the
                image frame.

        Returns:
            (np.ndarray): 
                ROI_images_rs (np.ndarray): 
                    The resized ROI images.
        """        
        scale_forRS = 0.7 * um_per_pixel  ## hardcoded for now sorry
        return np.stack([resize_affine(img, scale=scale_forRS, clamp_range=True) for img in ROI_images], axis=0)


    def generate_latents(self) -> torch.Tensor:
        """
        Passes the data in the dataloader through the network and generates latents.

        Returns:
            (torch.Tensor): 
                latents (torch.Tensor): 
                    Latents for each ROI (Region of Interest).
        """
        if hasattr(self, 'dataloader') == False:
            raise Exception('dataloader not defined. Call generate_dataloader() first.')

        print(f'starting: running data through network')
        self.latents = torch.cat([self.net(data[0][0].to(self._device)).detach() for data in tqdm(self.dataloader, mininterval=5)], dim=0).cpu()
        print(f'completed: running data through network')

        gc.collect()
        torch.cuda.empty_cache()
        gc.collect()
        torch.cuda.empty_cache()
        
        return self.latents


###################################
########### RESIZING ##############
###################################


def resize_affine(
    img: np.ndarray, 
    scale: float, 
    clamp_range: bool = False,
) -> np.ndarray:
    """
    Resizes an image using an affine transformation, scaled by a factor.

    Args:
        img (np.ndarray): 
            The input image to resize. Shape: *(H, W)*
        scale (float): 
            The scale factor to apply for resizing.
        clamp_range (bool): 
            If ``True``, the image will be clamped to the range [min(img),
            max(img)] to prevent interpolation from extending outside of the
            image's range. (Default is ``False``)

    Returns:
        (np.ndarray): 
            resized_image (np.ndarray): 
                The resized image.
    """
    img_rs = np.array(torchvision.transforms.functional.affine(
        img=PIL.Image.fromarray(img),
        angle=0, translate=[0,0], shear=0,
        scale=scale,
        interpolation=torchvision.transforms.InterpolationMode.BICUBIC
    ))

    if clamp_range:
        clamp_high = img.max()
        clamp_low = img.min()

        img_rs[img_rs>clamp_high] = clamp_high
        img_rs[img_rs<clamp_low] = clamp_low

    return img_rs

def resize_images(
    imgs: np.ndarray,
    scale: float,
    clamp_range: bool = False,
) -> np.ndarray:
    """
    Resizes images using an affine transformation, scaled by a factor.
    Uses torch.nn.functional.grid_sample to perform the resizing.
    
    Args:
        imgs (np.ndarray): 
            The input images to resize. Shape: *(N, H, W)*
        scale (float): 
            The scale factor to apply for resizing.
        clamp_range (bool): 
            If ``True``, the image will be clamped to the range [min(img),
            max(img)] to prevent interpolation from extending outside of the
            image's range. (Default is ``False``)

    Returns:
        (np.ndarray): 
            resized_images (np.ndarray): 
                The resized images. Shape: *(N, H, W)*
    """
    imgs = imgs[None, ...] if imgs.ndim == 2 else imgs
    imgs_rs = img_size = imgs.shape[1:]

    meshgrid_out = torch.stack(torch.meshgrid(torch.linspace(-1, 1, img_size[0]), torch.linspace(-1, 1, img_size[1]), indexing='xy'), dim=-1)

    imgs_rs = torch.nn.functional.grid_sample(
        input=torch.as_tensor(imgs)[None, ...],
        grid=meshgrid_out[None, ...] / scale,
        mode='bicubic',
        padding_mode='zeros',
        align_corners=True,
    )[0].numpy()

    if clamp_range:
        imgs_rs = np.clip(imgs_rs, a_min=imgs.min(axis=(1,2), keepdims=True), a_max=imgs.max(axis=(1,2), keepdims=True))

    return imgs_rs

def resize_affine2(
    imgs: np.ndarray, 
    scale: float, 
    clamp_range: bool = False,
) -> np.ndarray:
    """
    Resizes an image using an affine transformation, scaled by a factor.

    Args:
        img (np.ndarray): 
            The input images to resize. Shape: *(N, H, W)*
        scale (float): 
            The scale factor to apply for resizing.
        clamp_range (bool): 
            If ``True``, the image will be clamped to the range [min(img),
            max(img)] to prevent interpolation from extending outside of the
            image's range. (Default is ``False``)

    Returns:
        (np.ndarray): 
            resized_image (np.ndarray): 
                The resized image.
    """
    img_rs = np.array(torchvision.transforms.functional.affine(
        img=PIL.Image.fromarray(imgs.transpose(1,2,0)),
        angle=0, translate=[0,0], shear=0,
        scale=scale,
        interpolation=torchvision.transforms.InterpolationMode.BICUBIC
    )).transpose(2,0,1)

    if clamp_range:
        imgs_rs = np.clip(imgs_rs, a_min=imgs.min(axis=(1,2), keepdims=True), a_max=imgs.max(axis=(1,2), keepdims=True))

    return img_rs


###################################
########### FROM GRC ##############
###################################

class TileChannels(Module):
    """
    Expand dimension dim in X_in and tile to be N channels.
    RH 2021
    """
    def __init__(self, dim=0, n_channels=3):
        """
        Initializes the class.
        Args:
            dim (int):
                The dimension to tile.
            n_channels (int):
                The number of channels to tile to.
        """
        super().__init__()
        self.dim = dim
        self.n_channels = n_channels

    def forward(self, tensor):
        dims = [1]*len(tensor.shape)
        dims[self.dim] = self.n_channels
        return torch.tile(tensor, dims)
    def __repr__(self):
        return f"TileChannels(dim={self.dim})"

class Unsqueeze(Module):
    """
    Expand dimension dim in X_in and tile to be N channels.
    JZ 2023
    """
    def __init__(self, dim=0):
        """
        Initializes the class.
        Args:
            dim (int):
                The dimension to tile.
            n_channels (int):
                The number of channels to tile to.
        """
        super().__init__()
        self.dim = dim

    def forward(self, tensor):
        return torch.unsqueeze(tensor, self.dim)
    def __repr__(self):
        return f"Unsqueeze(dim={self.dim})"

class ScaleDynamicRange(Module):
    """
    Min-max scaling of the input tensor.
    RH 2021
    """
    def __init__(self, scaler_bounds=(0,1), epsilon=1e-9):
        """
        Initializes the class.
        Args:
            scaler_bounds (tuple):
                The bounds of how much to multiply the image by
                 prior to adding the Poisson noise.
             epsilon (float):
                 Value to add to the denominator when normalizing.
        """
        super().__init__()

        self.bounds = scaler_bounds
        self.range = scaler_bounds[1] - scaler_bounds[0]

        self.epsilon = epsilon

    def forward(self, tensor):
        tensor_minSub = tensor - tensor.min()
        return tensor_minSub * (self.range / (tensor_minSub.max()+self.epsilon))
    def __repr__(self):
        return f"ScaleDynamicRange(scaler_bounds={self.bounds})"


class dataset_simCLR(Dataset):
    """
    Args:
        X (Union[torch.Tensor, np.array, List[float]]): 
            Images. Expected shape: *(n_samples, height, width)*. Currently
            expects no channel dimension. If/when it exists, then shape should
            be *(n_samples, n_channels, height, width)*.
        y (Union[torch.Tensor, np.array, List[int]]): 
            Labels. Shape: *(n_samples)*.
        n_transforms (int): 
            Number of transformations to apply to each image. Should be >= 1.
            (Default is ``2``)
        transform (Optional[Callable]): 
            Optional transform to be applied on a sample. See
            torchvision.transforms for more information. Can use
            torch.nn.Sequential(a, bunch, of, transforms,) or other methods
            from torchvision.transforms. \n
            * If not ``None``: Transform(s) are applied to each image and the
              output shape of X_sample_transformed for __getitem__ will be
              *(n_samples, n_transforms, n_channels, height, width)*.
            * If ``None``: No transform is applied and output shape of
              X_sample_trasformed for __getitem__ will be *(n_samples,
              n_channels, height, width)* (which is missing the n_transforms
              dimension). \n
            (Default is ``None``)
        DEVICE (str): 
            Device on which the data will be stored and transformed. Best to
            leave this as 'cpu' and do .to(DEVICE) on the data for the training
            loop. (Default is ``'cpu'``)
        dtype_X (torch.dtype): 
            Data type of X. (Default is ``torch.float32``)
        dtype_y (torch.dtype): 
            Data type of y. (Default is ``torch.int64``)
        temp_uncetainty (float):
            Temperture term applied to the CrossEntropyLoss input. (Default is
            ``1.0`` for no change)

    Example:
        .. highlight:: python
        .. code-block:: python

            transforms = torch.nn.Sequential(
                torchvision.transforms.RandomHorizontalFlip(p=0.5),

                torchvision.transforms.GaussianBlur(
                    5,
                    sigma=(0.01, 1.)
                ),
                torchvision.transforms.RandomPerspective(
                    distortion_scale=0.6,
                    p=1,
                    interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
                    fill=0
                ),
                torchvision.transforms.RandomAffine(
                    degrees=(-180,180),
                    translate=(0.4, 0.4),
                    scale=(0.7, 1.7),
                    shear=(-20, 20, -20, 20),
                    interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
                    fill=0,
                    fillcolor=None,
                    resample=None
                ),
            )
            scripted_transforms = torch.jit.script(transforms)

            dataset = dataset_simCLR(  torch.tensor(images),
                                        labels,
                                        n_transforms=2,
                                        transform=scripted_transforms,
                                        DEVICE='cpu',
                                        dtype_X=torch.float32,
                                        dtype_y=torch.int64)

            dataloader = torch.utils.data.DataLoader(   dataset,
                                                    batch_size=64,
                                                    shuffle=True,
                                                    drop_last=True,
                                                    pin_memory=False,
                                                    num_workers=0)
    """
    def __init__(   
        self,
        X: Union[torch.Tensor, np.array, List[float]],
        y: Union[torch.Tensor, np.array, List[int]],
        n_transforms: int = 2,
        transform: Optional[Callable] = None,
        DEVICE: str = 'cpu',
        dtype_X: torch.dtype = torch.float32,
        dtype_y: torch.dtype = torch.int64,
    ):
        """
        Initializes the dataset_simCLR object with the given images, labels, and
        optional settings.
        """

        self.X = torch.as_tensor(X, dtype=dtype_X, device=DEVICE) # first dim will be subsampled from. Shape: (n_samples, n_channels, height, width)
        self.X = self.X[:,None,...]
        self.y = torch.as_tensor(y, dtype=dtype_y, device=DEVICE) # first dim will be subsampled from.

        self.idx = torch.arange(self.X.shape[0], device=DEVICE)
        self.n_samples = self.X.shape[0]

        self.transform = transform
        self.n_transforms = n_transforms

        if X.shape[0] != y.shape[0]:
            raise ValueError('RH Error: X and y must have same first dimension shape')

    def tile_channels(
        self,
        X_in: Union[torch.Tensor, np.ndarray],
        dim: int = -3,
    ) -> Union[torch.Tensor, np.ndarray]:
        """
        Expand dimension dim in X_in and tile to be 3 channels.

        Args:
            X_in (torch.Tensor or np.ndarray): 
                Input image with shape: *(n_channels==1, height, width)*
            dim (int): 
                Dimension to expand. (Default is ``-3``)

        Returns:
            (torch.Tensor or np.ndarray): 
                X_out (torch.Tensor or np.ndarray):
                    Output image with shape: *(n_channels==3, height, width)*
        """
        dims = [1]*len(X_in.shape)
        dims[dim] = 3
        return torch.tile(X_in, dims)
    
    def __len__(self):
        """
        Get the total number of samples in the dataset.

        Returns:
            (int): 
                n_samples (int): 
                    The total number of samples.
        """
        return self.n_samples
    
    def __getitem__(
        self,
        idx: int,
    ) -> Tuple[Union[torch.Tensor, np.ndarray], int, int, int]:
        """
        Retrieves and transforms a sample.

        Args:
            idx (int): 
                Index of the sample to retrieve.

        Returns:
            (Tuple): tuple containing:
                X_sample_transformed (torch.Tensor or np.ndarray):
                    Transformed sample(s). Shape: 
                        * If transform is ``None``: *(batch_size, n_channels, height, width)*
                        * If transform is not ``None``: *(n_transforms, batch_size, n_channels, height, width)*
                y_sample (int):
                    Label of the sample.
                idx_sample (int):
                    Index of the sample.
                sample_weight (int):
                    Weight of the sample. Always 1.
        """
        y_sample = self.y[idx]
        idx_sample = self.idx[idx]

        sample_weight = 1

        X_sample_transformed = []
        if self.transform is not None:
            for ii in range(self.n_transforms):
                X_transformed = self.transform(self.X[idx_sample])
                X_sample_transformed.append(X_transformed)
        else:
            X_sample_transformed = self.tile_channels(self.X[idx_sample], dim=-3)

        return X_sample_transformed, y_sample, idx_sample, sample_weight
    


## Import basic libraries
from pathlib import Path
import copy
import tempfile
from IPython.display import display
import time

# import matplotlib.pyplot as plt
import numpy as np

## Import roicat submodules
from . import data_importing, ROInet, helpers, util, visualization, tracking, classification

def pipeline_tracking(params: dict):
    """
    Pipeline for tracking ROIs across sessions.
    RH 2023

    Args:
        params (dict):
            Dictionary of parameters. See
            ``roicat.util.get_default_parameters(pipeline='tracking')`` for
            details.

    Returns:
        (tuple): tuple containing:
            results (dict):
                Dictionary of results.
            run_data (dict):
                Dictionary containing the different class objects used in the
                pipeline.
            params (dict):
                Parameters used in the pipeline. See
                ``roicat.helpers.prepare_params()`` for details.
    """
    ## Start timer
    tic_start = time.time()

    ## Prepare params
    defaults = util.get_default_parameters(pipeline='tracking')
    params = helpers.prepare_params(params, defaults, verbose=True)
    display(params)

    ## Prepare state variables
    VERBOSE = params['general']['verbose']
    DEVICE = helpers.set_device(use_GPU=params['general']['use_GPU'])
    SEED = _set_random_seed(
        seed=params['general']['random_seed'],
        deterministic=params['general']['random_seed'] is not None,
    )

    if params['data_loading']['data_kind'] == 'data_suite2p':
        assert params['data_loading']['dir_outer'] is not None, f"params['data_loading']['dir_outer'] must be specified if params['data_loading']['data_kind'] is 'data_suite2p'."
        paths_allStat = helpers.find_paths(
            dir_outer=params['data_loading']['dir_outer'],
            reMatch='stat.npy',
            depth=4,
            find_files=True,
            find_folders=False,
            natsorted=True,
        )[:]
        paths_allOps  = [str(Path(path).resolve().parent / 'ops.npy') for path in paths_allStat][:]

        print(f"Found the following stat.npy files:")
        [print(f"    {path}") for path in paths_allStat]
        print(f"Found the following corresponding ops.npy files:")
        [print(f"    {path}") for path in paths_allOps]


        ## Import data
        data = data_importing.Data_suite2p(
            paths_statFiles=paths_allStat[:],
            paths_opsFiles=paths_allOps[:],
            verbose=VERBOSE,
            **{**params['data_loading']['common'], **params['data_loading']['data_suite2p']},
        )
        assert data.check_completeness(verbose=False)['tracking'], f"Data object is missing attributes necessary for tracking."
    elif params['data_loading']['data_kind'] == 'roicat':
        paths_allDataObjs = helpers.find_paths(
            dir_outer=params['data_loading']['dir_outer'],
            reMatch=params['data_loading']['data_roicat']['filename_search'],
            depth=1,
            find_files=False,
            find_folders=True,
            natsorted=True,
        )[:]
        assert len(paths_allDataObjs) == 1, f"ERROR: Found {len(paths_allDataObjs)} files matching the search pattern '{params['data_loading']['data_roicat']['filename_search']}' in '{params['data_loading']['dir_outer']}'. Exactly one file must be found."
        
        data = data_importing.Data_roicat()
        # data.load(path_load=paths_allDataObjs[0])

        data.import_from_dict(
            dict_load=util.RichFile_ROICaT(path=paths_allDataObjs[0]).load(),
        )
    else:
        raise NotImplementedError(f"params['data_loading']['data_kind'] == '{params['data_loading']['data_kind']}' is not yet implemented.")

    ## Alignment
    aligner = tracking.alignment.Aligner(verbose=True)
    FOV_images = aligner.augment_FOV_images(
        FOV_images=data.FOV_images,
        spatialFootprints=data.spatialFootprints,
        **params['alignment']['augment'],
    )
    aligner.fit_geometric(
        ims_moving=FOV_images,  ## input images
        **params['alignment']['fit_geometric'],
    )
    aligner.transform_images_geometric(FOV_images);
    aligner.fit_nonrigid(
        ims_moving=aligner.ims_registered_geo,  ## Input images. Typically the geometrically registered images
        remappingIdx_init=aligner.remappingIdx_geo,  ## The remappingIdx between the original images (and ROIs) and ims_moving
        **params['alignment']['fit_nonrigid'],
    )
    aligner.transform_images_nonrigid(FOV_images);
    aligner.transform_ROIs(
        ROIs=data.spatialFootprints, 
        remappingIdx=aligner.remappingIdx_nonrigid,
        **params['alignment']['transform_ROIs'],
    );


    ## Blur ROIs
    blurrer = tracking.blurring.ROI_Blurrer(
        frame_shape=(data.FOV_height, data.FOV_width),  ## FOV height and width
        plot_kernel=False,  ## Whether to visualize the 2D gaussian
        **params['blurring'],
    )
    blurrer.blur_ROIs(
        spatialFootprints=aligner.ROIs_aligned[:],
    )


    ## ROInet embedding
    dir_temp = tempfile.gettempdir()

    roinet = ROInet.ROInet_embedder(
        device=DEVICE,  ## Which torch device to use ('cpu', 'cuda', etc.)
        dir_networkFiles=dir_temp,  ## Directory to download the pretrained network to
        verbose=VERBOSE,  ## Whether to print updates
        **params['ROInet']['network'],
    )
    roinet.generate_dataloader(
        ROI_images=data.ROI_images,  ## Input images of ROIs
        um_per_pixel=data.um_per_pixel,  ## Resolution of FOV
        pref_plot=False,  ## Whether or not to plot the ROI sizes
        **params['ROInet']['dataloader'],
    );
    roinet.generate_latents();


    ## Scattering wavelet embedding
    swt = tracking.scatteringWaveletTransformer.SWT(
        image_shape=data.ROI_images[0].shape[1:3],  ## size of a cropped ROI image
        device=DEVICE,  ## PyTorch device
        kwargs_Scattering2D=params['SWT']['kwargs_Scattering2D'],
    )
    swt.transform(
        ROI_images=roinet.ROI_images_rs,  ## All the cropped and resized ROI images
        batch_size=params['SWT']['batch_size'],
    );


    ## Compute similarities
    sim = tracking.similarity_graph.ROI_graph(
        frame_height=data.FOV_height,
        frame_width=data.FOV_width,
        verbose=VERBOSE,  ## Whether to print outputs
        **params['similarity_graph']['sparsification']
    )
    s_sf, s_NN, s_SWT, s_sesh = sim.compute_similarity_blockwise(
        spatialFootprints=blurrer.ROIs_blurred,  ## Mask spatial footprints
        features_NN=roinet.latents,  ## ROInet output latents
        features_SWT=swt.latents,  ## Scattering wavelet transform output latents
        ROI_session_bool=data.session_bool,  ## Boolean array of which ROIs belong to which sessions
    #     spatialFootprint_maskPower=1.0,  ##  An exponent to raise the spatial footprints to to care more or less about bright pixels
        **params['similarity_graph']['compute_similarity'],
    );
    sim.make_normalized_similarities(
        centers_of_mass=data.centroids,  ## ROI centroid positions
        features_NN=roinet.latents,  ## ROInet latents
        features_SWT=swt.latents,  ## SWT latents
        device=DEVICE,    
        k_max=data.n_sessions * params['similarity_graph']['normalization']['k_max'],
        k_min=data.n_sessions * params['similarity_graph']['normalization']['k_min'],
        algo_NN=params['similarity_graph']['normalization']['algo_NN'],
    )


    ## Clustering
    clusterer = tracking.clustering.Clusterer(
        s_sf=sim.s_sf,
        s_NN_z=sim.s_NN_z,
        s_SWT_z=sim.s_SWT_z,
        s_sesh=sim.s_sesh,
        verbose=VERBOSE,
    )
    if params['clustering']['mixing_method'] == 'automatic':
        kwargs_makeConjunctiveDistanceMatrix_best = clusterer.find_optimal_parameters_for_pruning(
            seed=SEED,
            **params['clustering']['parameters_automatic_mixing'],
        )
    elif params['clustering']['mixing_method'] == 'manual':
        kwargs_makeConjunctiveDistanceMatrix_best = params['clustering']['parameters_manual_mixing']
    else:
        ## Not implemented
        raise NotImplementedError(f"Mixing method '{params['clustering']['mixing_method']}' is not implemented. Select from: ['automatic', 'manual']")

    clusterer.make_pruned_similarity_graphs(
        kwargs_makeConjunctiveDistanceMatrix=kwargs_makeConjunctiveDistanceMatrix_best,
        **params['clustering']['pruning'],
    )

    def choose_clustering_method(method='automatic', n_sessions_switch=8, n_sessions=None):
        if method == 'automatic':
            method_out = 'hdbscan'.upper() if n_sessions >= n_sessions_switch else 'sequential_hungarian'.upper()
        else:
            method_out = method.upper()
        assert method_out.upper() in ['hdbscan'.upper(), 'sequential_hungarian'.upper()]
        return method_out
    method_clustering = choose_clustering_method(
        method=params['clustering']['cluster_method']['method'],
        n_sessions_switch=params['clustering']['cluster_method']['n_sessions_switch'],
        n_sessions=data.n_sessions,
    )

    if method_clustering == 'hdbscan'.upper():
        labels = clusterer.fit(
            d_conj=clusterer.dConj_pruned,  ## Input distance matrix
            session_bool=data.session_bool,  ## Boolean array of which ROIs belong to which sessions
            **params['clustering']['hdbscan'],
        )
    elif method_clustering == 'sequential_hungarian'.upper():
        labels = clusterer.fit_sequentialHungarian(
            d_conj=clusterer.dConj_pruned,  ## Input distance matrix
            session_bool=data.session_bool,  ## Boolean array of which ROIs belong to which sessions
            **params['clustering']['sequential_hungarian'],
        )
    else:
        raise ValueError('Clustering method not recognized. This should never happen.')

    quality_metrics = clusterer.compute_quality_metrics();

    ## Collect results
    labels_squeezed, labels_bySession, labels_bool, labels_bool_bySession, labels_dict = tracking.clustering.make_label_variants(labels=labels, n_roi_bySession=data.n_roi)

    results_clusters = {
        'labels': labels_squeezed,
        'labels_bySession': labels_bySession,
        'labels_dict': labels_dict,
        'quality_metrics': quality_metrics,
    }

    results_all = {
        "clusters":{
            "labels": util.JSON_List(labels_squeezed),
            "labels_bySession": util.JSON_List(labels_bySession),
            "labels_bool": labels_bool,
            "labels_bool_bySession": labels_bool_bySession,
            "labels_dict": util.JSON_Dict(labels_dict),
            "quality_metrics": util.JSON_Dict(clusterer.quality_metrics) if hasattr(clusterer, 'quality_metrics') else None,
        },
        "ROIs": {
            "ROIs_aligned": aligner.ROIs_aligned,
            "ROIs_raw": data.spatialFootprints,
            "frame_height": data.FOV_height,
            "frame_width": data.FOV_width,
            "idx_roi_session": np.where(data.session_bool)[1],
            "n_sessions": data.n_sessions,
        },
        "input_data": {
            "paths_stat": data.paths_stat if hasattr(data, 'paths_stat') else None,
            "paths_ops": data.paths_ops if hasattr(data, 'paths_ops') else None,
        },
    }

    run_data = {
        'data': data.__dict__,
        'aligner': aligner.__dict__,
        'blurrer': blurrer.__dict__,
        'roinet': roinet.__dict__,
        'swt': swt.__dict__,
        'sim': sim.__dict__,
        'clusterer': clusterer.__dict__,
    }
    params_used = {name: mod['params'] for name, mod in run_data.items()}

    ## Print some results
    print(f'Number of clusters: {len(np.unique(results_clusters["labels"]))}')
    print(f'Number of discarded ROIs: {(np.array(results_clusters["labels"])==-1).sum()}')

    ## Save results
    if params['results_saving']['dir_save'] is not None:

        dir_save = Path(params['results_saving']['dir_save']).resolve()
        name_save = str(params['results_saving']['prefix_name_save'])

        print(f'dir_save: {dir_save}')

        paths_save = {
            'results_clusters': str(Path(dir_save) / f'{name_save}.tracking.results_clusters.json'),
            'params_used':      str(Path(dir_save) / f'{name_save}.tracking.params_used.json'),
            'results_all':      str(Path(dir_save) / f'{name_save}.tracking.results_all.richfile'),
            'run_data':         str(Path(dir_save) / f'{name_save}.tracking.run_data.richfile'),
        }

        helpers.json_save(obj=results_clusters, filepath=paths_save['results_clusters']);
        helpers.json_save(obj=params_used, filepath=paths_save['params_used']);
        util.RichFile_ROICaT(path=paths_save['results_all']).save(obj=results_all, overwrite=True);
        util.RichFile_ROICaT(path=paths_save['run_data']).save(obj=run_data, overwrite=True);

    
        ## Visualize results
        ### Save some figures
        
        #### Save FOV_images as .png files
        def save_image(array, path, normalize=True):
            ## Use PIL to save the image
            from PIL import Image
            Path(path).parent.mkdir(parents=True, exist_ok=True)
            Image.fromarray((np.array(array / array.max() if normalize else array) * 255).astype(np.uint8)).save(path)
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'FOV_images' / f'FOV_images_{ii}.png') ) for ii, array in enumerate(data.FOV_images)]
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'FOV_images_aligned_geometric' / f'FOV_images_aligned_geometric_{ii}.png') ) for ii, array in enumerate(aligner.ims_registered_geo)]
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'FOV_images_aligned_nonrigid' / f'FOV_images_aligned_nonrigid_{ii}.png') ) for ii, array in enumerate(aligner.ims_registered_nonrigid)]
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'ROIs' / f'ROIs_{ii}.png') ) for ii, array in enumerate(data.get_maxIntensityProjection_spatialFootprints())]
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'ROIs_aligned' / f'ROIs_aligned_{ii}.png') ) for ii, array in enumerate(aligner.get_ROIsAligned_maxIntensityProjection(normalize=True))]
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'ROIs_aligned_blurred' / f'ROIs_aligned_blurred_{ii}.png') ) for ii, array in enumerate(blurrer.get_ROIsBlurred_maxIntensityProjection())]
        
        #### Save some sample ROI images
        [save_image(array, str(Path(dir_save).resolve() / 'visualization' / 'ROIs_sample' / f'ROIs_sample_{ii}.png') ) for ii, array in enumerate(roinet.ROI_images_rs[:100])]
        
        #### Save the similarity graphy blocks
        fig = sim.visualize_blocks()
        (Path(dir_save).resolve() / 'visualization' / 'similarity_graph').mkdir(parents=True, exist_ok=True)
        fig.savefig(str(Path(dir_save).resolve() / 'visualization' / 'similarity_graph' / 'blocks.png'))
        
        #### Save the similarity / distance plots for the given conjunctive distance matrix kwargs
        fig = clusterer.plot_distSame(kwargs_makeConjunctiveDistanceMatrix=kwargs_makeConjunctiveDistanceMatrix_best)
        (Path(dir_save).resolve() / 'visualization' / 'clustering').mkdir(parents=True, exist_ok=True)
        fig.savefig(str(Path(dir_save).resolve() / 'visualization' / 'clustering' / 'dist.png'))
        fig, axs = clusterer.plot_similarity_relationships(
            plots_to_show=[1,2,3], 
            max_samples=100000,  ## Make smaller if it is running too slow
            kwargs_scatter={'s':1, 'alpha':0.2},
            kwargs_makeConjunctiveDistanceMatrix=kwargs_makeConjunctiveDistanceMatrix_best
        )
        (Path(dir_save).resolve() / 'visualization' / 'clustering').mkdir(parents=True, exist_ok=True)
        fig.savefig(str(Path(dir_save).resolve() / 'visualization' / 'clustering' / 'similarity_relationships.png'))
        
        #### Save the clustering results
        fig, axs = tracking.clustering.plot_quality_metrics(
            quality_metrics=quality_metrics, 
            labels=labels_squeezed, 
            n_sessions=data.n_sessions,
        )
        (Path(dir_save).resolve() / 'visualization' / 'clustering').mkdir(parents=True, exist_ok=True)
        fig.savefig(str(Path(dir_save).resolve() / 'visualization' / 'clustering' / 'quality_metrics.png'))
        
        ### Save a gif of the ROIs
        FOV_clusters = visualization.compute_colored_FOV(
            spatialFootprints=[r.power(1.0) for r in results_all['ROIs']['ROIs_aligned']],  ## Spatial footprint sparse arrays
            FOV_height=results_all['ROIs']['frame_height'],
            FOV_width=results_all['ROIs']['frame_width'],
            labels=results_all["clusters"]["labels_bySession"],  ## cluster labels
        #     labels=(np.array(results["clusters"]["labels"])!=-1).astype(np.int64),  ## cluster labels
        #     alphas_labels=confidence*1.5,  ## Set brightness of each cluster based on some 1-D array
        #     alphas_labels=(clusterer.quality_metrics['cluster_silhouette'] > 0) * (clusterer.quality_metrics['cluster_intra_means'] > 0.4),
        #     alphas_sf=clusterer.quality_metrics['sample_silhouette'],  ## Set brightness of each ROI based on some 1-D array
        )
        helpers.save_gif(
            array=helpers.add_text_to_images(
                images=[(f * 255).astype(np.uint8) for f in FOV_clusters], 
                text=[[f"{ii}",] for ii in range(len(FOV_clusters))], 
                font_size=3,
                line_width=10,
                position=(30, 90),
            ), 
            path=str(Path(dir_save).resolve() / 'visualization' / 'FOV_clusters.gif'),
            frameRate=10.0,
            loop=0,
        )



    ## End timer
    tic_end = time.time()
    print(f"Elapsed time: {tic_end - tic_start:.2f} seconds")
    
    return results_all, run_data, params

def _set_random_seed(seed=None, deterministic=False):
    """
    Set random seed for reproducibility.
    RH 2023

    Args:
        seed (int, optional):
            Random seed.
            If None, a random seed (spanning int32 integer range) is generated.
        deterministic (bool, optional):
            Whether to make packages deterministic.

    Returns:
        (int):
            seed (int):
                Random seed.
    """
    ### random seed (note that optuna requires a random seed to be set within the pipeline)
    import numpy as np
    seed = int(np.random.randint(0, 2**31 - 1, dtype=np.uint32)) if seed is None else seed

    np.random.seed(seed)
    import torch
    torch.manual_seed(seed)
    import random
    random.seed(seed)
    import cv2
    cv2.setRNGSeed(seed)

    ## Make torch deterministic
    torch.use_deterministic_algorithms(deterministic)
    ## Make cudnn deterministic
    torch.backends.cudnn.deterministic = deterministic
    torch.backends.cudnn.benchmark = not deterministic
    
    return seed


# Imports
import sys
import os
import numpy as np
import torch
import torchvision
from PIL import Image
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.decomposition import PCA
from functools import partial
from typing import Optional, List, Tuple, Union, Dict, Any

from torch.nn import CrossEntropyLoss
from torch.optim import Adam

from .. import helpers

class ModelTackOn(torch.nn.Module):
    """
    Class to attach fully connected layers to the end of a pretrained
    network to create a SimCLR model with "head" and "latent" outputs.
    JZ, RH 2021-2023

    Args:
        base_model (torch.nn.Module):
            Pretrained model to which fully connected layers will be attached
        un_modified_model (torch.nn.Module):
            Pretrained model that has not been modified
        data_dim (tuple):
            Dimensions of the data to be passed through the model
        pre_head_fc_sizes (list):
            List of fully connected layer sizes to be attached before the head
        post_head_fc_sizes (list):
            List of fully connected layer sizes to be attached after the head
        nonlinearity (str):
            Nonlinearity to be used in the fully connected layers
        kwargs_nonlinearity (dict):
            Keyword arguments to be passed to the nonlinearity function
        non_singular_pca_size (int):
            Size of the PCA to be used to create the latent space
            NOT YET IMPLEMENTED
            
    Returns:
        model (torch.nn.Module):
            Model with fully connected layers attached
    """

    def __init__(
        self, 
        base_model: torch.nn.Module, 
        un_modified_model: torch.nn.Module,
        data_dim: Tuple[int, int, int, int]=(1,3,36,36),
        pre_head_fc_sizes: List[int]=[100],
        post_head_fc_sizes: List[int]=[100], 
        nonlinearity: str='relu', 
        kwargs_nonlinearity={},
        non_singular_pca_size=None,
    ):
        super(ModelTackOn, self).__init__()
        self.base_model = base_model
        final_base_layer = list(un_modified_model.children())[-1]
        
        self.data_dim = data_dim
        self.non_singular_pca_size = non_singular_pca_size

        self.pre_head_fc_lst = []
        self.post_head_fc_lst = []
            
        self.nonlinearity = nonlinearity
        self.kwargs_nonlinearity = kwargs_nonlinearity

        self.init_prehead(pre_head_fc_sizes)
        self.init_posthead(pre_head_fc_sizes[-1], post_head_fc_sizes)
        # self.init_pca_layer(pre_head_fc_sizes[-1])
    
    def init_prehead(self, pre_head_fc_sizes):
        """
        Initialize the fully connected layers to be attached before the head
        
        Args:
            pre_head_fc_sizes (list):
                List of fully connected layer sizes to be attached before the head
        """
        for i, pre_head_fc in enumerate(pre_head_fc_sizes):
            if i == 0:
                in_features = self.base_model(torch.rand(*(self.data_dim))).data.squeeze().shape[0]  ## RH EDIT
            else:
                in_features = pre_head_fc_sizes[i - 1]
            fc_layer = torch.nn.Linear(in_features=in_features, out_features=pre_head_fc)
            self.add_module(f'PreHead_{i}', fc_layer)
            self.pre_head_fc_lst.append(fc_layer)

            non_linearity = torch.nn.__dict__[self.nonlinearity](**self.kwargs_nonlinearity)
            self.add_module(f'PreHead_{i}_NonLinearity', non_linearity)
            self.pre_head_fc_lst.append(non_linearity)

    def init_posthead(self, prv_size, post_head_fc_sizes):
        """
        Initialize the fully connected layers to be attached after the head

        Args:
            prv_size (int):
                Size of the final layer of the base model
            post_head_fc_sizes (list):
                List of fully connected layer sizes to be attached after the head
        """
        for i, post_head_fc in enumerate(post_head_fc_sizes):
            if i == 0:
                in_features = prv_size
            else:
                in_features = post_head_fc_sizes[i - 1]
            fc_layer = torch.nn.Linear(in_features=in_features, out_features=post_head_fc)
            self.add_module(f'PostHead_{i}', fc_layer)
            self.post_head_fc_lst.append(fc_layer)

            non_linearity = torch.nn.__dict__[self.nonlinearity](**self.kwargs_nonlinearity)    
            self.add_module(f'PostHead_{i}_NonLinearity', non_linearity)
            self.pre_head_fc_lst.append(non_linearity)
    
    # def init_pca_layer(self, pca_size):
    #     """
    #     Initialize the PCA layer with identity weights and biases

    #     Args:
    #         pca_size (int):
    #             Size of the PCA to be used to create the latent space
    #     """
    #     self.pca_layer = torch.nn.Sequential(
    #         torch.nn.Linear(pca_size, pca_size),
    #         torch.nn.Linear(pca_size, pca_size, bias=False)
    #     )
    #     self.pca_layer[0].weight = torch.nn.Parameter(torch.tensor(np.eye(pca_size,),dtype=torch.float32))
    #     self.pca_layer[0].bias = torch.nn.Parameter(torch.tensor(np.zeros(pca_size,),dtype=torch.float32))
    #     self.pca_layer[1].weight = torch.nn.Parameter(torch.tensor(np.eye(pca_size,),dtype=torch.float32))
    #     self.pca_layer[1].bias = torch.nn.Parameter(torch.tensor(np.zeros(pca_size,),dtype=torch.float32))

    #     self.add_module(f'PCA_Layer', self.pca_layer)
    
    def forward_latent(self, X):
        """
        Run the model forward to get the latent representation of the data
        (final output of model—used for similarity calculations in SimCLR training)

        Args:
            X (torch.Tensor):
                Input data to be run through the model

        Returns:
            latent (torch.Tensor):
                Latent representation of the input data
        """
        interim = self.base_model(X)
        head = self.get_head(interim)
        latent = self.get_latent(head)
        return latent

    def forward_head(self, X):
        """
        Run the model forward to get the head output of the model (should be used for training PCA)

        Args:
            X (torch.Tensor):
                Input data to be run through the model

        Returns:
            latent (torch.Tensor):
                Latent representation of the input data
        """
        interim = self.base_model(X)
        head = self.get_head(interim)
        return head

    # def forward_head_pca(self, X):
    #     """
    #     Run the model forward to get the head output of the model, passed through a pre-fit PCA layer
    #     (used for classification)

    #     Args:
    #         X (torch.Tensor):
    #             Input data to be run through the model

    #     Returns:
    #         head_pca (torch.Tensor):
    #             Head output of the model, passed through a pre-fit PCA layer
    #     """
    #     interim = self.base_model(X)
    #     head = self.get_head(interim)
    #     head_pca = self.pca_layer(head)[...,:self.non_singular_pca_size] if self.non_singular_pca_size is not None else self.pca_layer(head)
    #     return head_pca

    def get_head(self, base_out):
        """
        Run the model forward through the FC layers between the base model
        and the head output

        Args:
            base_out (torch.Tensor):
                Output of the base model

        Returns:
            head (torch.Tensor):
                Output of the FC layers (the head output used for classification)
        """
        interim = base_out
        for pre_head_layer in self.pre_head_fc_lst:
            interim = pre_head_layer(interim)
        head = interim
        return head

    def get_latent(self, head):
        """
        Run the model forward through the FC layers between the head output
        and the latent representation

        Args:
            head (torch.Tensor):
                Output of the FC layers (the head output used for classification)

        Returns:
            latent (torch.Tensor):
                Latent representation of the input data (used for SimCLR similarity training)
        """
        interim = head
        for post_head_layer in self.post_head_fc_lst:
            interim = post_head_layer(interim)
        latent = interim
        return latent

    # def set_pca_head_grad(self, requires_grad=False):
    #     """
    #     Set the gradient requirements for the PCA output head layers built on the head

    #     Args:
    #         requires_grad (bool):
    #             Whether or not to require gradients for the FC layers
    #     """
    #     for param in self.pca_layer.parameters():
    #         param.requires_grad = requires_grad
    
    def set_pre_head_grad(self, requires_grad=True):
        """
        Set the gradient requirements for the FC layers between the base model
        and the head output

        Args:
            requires_grad (bool):
                Whether or not to require gradients for the FC layers
        """
        for layer in self.pre_head_fc_lst:
            for param in layer.parameters():
                param.requires_grad = requires_grad
                
    def set_post_head_grad(self, requires_grad=True):
        """
        Set the gradient requirements for the FC layers between the head output
        and the latent representation

        Args:
            requires_grad (bool):
                Whether or not to require gradients for the FC layers
        """
        for layer in self.post_head_fc_lst:
            for param in layer.parameters():
                param.requires_grad = requires_grad

    def prep_contrast(self):
        """
        Set the gradient requirements for the FC layers between the base model
        and the head output and the FC layers between the head output
        and the latent representation to True

        Args:
            requires_grad (bool):
                Whether or not to require gradients for the FC layers
        """
        self.set_pre_head_grad(requires_grad=True)
        self.set_post_head_grad(requires_grad=True)
        # self.set_pca_head_grad(requires_grad=False)

    @property
    def device(self):
        """
        Get the device of the model

        Returns:
            device (torch.device):
                Device of the model
        """
        return next(self.parameters()).device

class Simclr_Model():
    """
    SimCLR model class

    Args:
        filepath_model_save (str):
            Filepath to which to save the model
        filepath_model_load (str):
            Filepath from which to load a pretrained model
        base_model (torch.nn.Module):
            Base torchvision model (or otherwise) to use for the SimCLR model
        head_pool_method (str):
            Pooling method to use for the head
        head_pool_method_kwargs (dict):
            Pooling method kwargs to use for the head  
        pre_head_fc_sizes (list):
            List of fully connected layer sizes to be attached before the head
        post_head_fc_sizes (list):
            List of fully connected layer sizes to be attached after the head
        head_nonlinearity (str):
            Nonlinearity to use after the FC layers
        head_nonlinearity_kwargs (dict):
            Nonlinearity kwargs to use after the FC layers
        block_to_unfreeze (str):
            Name of the block to unfreeze for training
        n_block_toInclude (int):
            Number of blocks to include in the base model
        image_out_size (int):
            Size of the output image (for resizing)
        forward_version (str):
            Version of the forward pass to use
    """
    def __init__(
            self,
            filepath_model_save: Optional[torch.nn.Module]=None, # Set filepath to load model
            filepath_model_load: Optional[torch.nn.Module]=None, # Set filepath to save model
            base_model: Optional[torch.nn.Module]=None, # Set base model to use
            head_pool_method: Optional[str]=None, # Set pooling method to use for the head
            head_pool_method_kwargs: Optional[dict]=None, # Set pooling method kwargs to use for the head
            pre_head_fc_sizes: Optional[list]=None, # Set the sizes of the FC layers to be attached before the head
            post_head_fc_sizes: Optional[int]=None, # Set the size of the FC layer to be attached after the head
            head_nonlinearity: Optional[str]=None, # Set the nonlinearity to use after the head
            head_nonlinearity_kwargs: Optional[dict]=None, # Set the nonlinearity kwargs to use after the head
            block_to_unfreeze: Optional[str]=None, # Unfreeze the model at and beyond the unfreeze_point
            n_block_toInclude: Optional[int]=None, # Set the number of blocks to include in the model
            image_out_size: Optional[int]=None, # Set the size of the output image
            forward_version: Optional[str]=None, # Set the version of the forward pass to use
            ):
        
        assert filepath_model_load is not None or (
            base_model is not None and
            head_pool_method is not None and
            head_pool_method_kwargs is not None and
            pre_head_fc_sizes is not None and
            post_head_fc_sizes is not None and
            head_nonlinearity is not None and
            head_nonlinearity_kwargs is not None and
            block_to_unfreeze is not None and
            n_block_toInclude is not None and
            image_out_size is not None and
            forward_version is not None
            ), "Either filepath_model_load or every other parameter must be set (except filepath_model_save)"

        # If loading model, load it from onnx, otherwise create one from scratch using the other parameters
        if filepath_model_load is not None:
            self.load_onnx(filepath_model_load)
        else:
            self.create_model(
                base_model=base_model,
                head_pool_method=head_pool_method,
                head_pool_method_kwargs=head_pool_method_kwargs,
                pre_head_fc_sizes=pre_head_fc_sizes,
                post_head_fc_sizes=post_head_fc_sizes,
                head_nonlinearity=head_nonlinearity,
                head_nonlinearity_kwargs=head_nonlinearity_kwargs,
                block_to_unfreeze=block_to_unfreeze,
                n_block_toInclude=n_block_toInclude,
                image_out_size=image_out_size,
                forward_version=forward_version
                )
        self.filepath_model_save = filepath_model_save
        self.filepath_model_load = filepath_model_load
            
    def create_model(
            self,
            base_model=None, # Freeze base_model
            head_pool_method=None,
            head_pool_method_kwargs=None,
            pre_head_fc_sizes=None,
            post_head_fc_sizes=None,
            head_nonlinearity=None,
            head_nonlinearity_kwargs=None,
            block_to_unfreeze=None, # Unfreeze the model at and beyond the unfreeze_point
            n_block_toInclude=None, # Unfreeze the model at and beyond the unfreeze_point
            image_out_size=None,
            forward_version=None
            ):
        """
        Create the model from scratch using the parameters

        Args:
            base_model (torch.nn.Module):
                Base torchvision model (or otherwise) to use for the SimCLR model
            head_pool_method (str):
                Pooling method to use for the head
            head_pool_method_kwargs (dict):
                Pooling method kwargs to use for the head
            pre_head_fc_sizes (list):
                List of fully connected layer sizes to be attached before the head  
            post_head_fc_sizes (list):
                List of fully connected layer sizes to be attached after the head
            head_nonlinearity (str):
                Nonlinearity to use after the FC layers
            head_nonlinearity_kwargs (dict):
                Nonlinearity kwargs to use after the FC layers
            block_to_unfreeze (str):
                Name of the block to unfreeze for training
            n_block_toInclude (int):
                Number of blocks to include in the base model
            image_out_size (int):
                Size of the output image (for resizing)
            forward_version (str):
                Version of the forward pass to use
        """

        base_model_frozen = base_model
        for param in base_model_frozen.parameters():
            param.requires_grad = False

        model_chopped = torch.nn.Sequential(list(base_model_frozen.children())[0][:n_block_toInclude])  ## 0.
        model_chopped_pooled = torch.nn.Sequential(model_chopped, torch.nn.__dict__[head_pool_method](**head_pool_method_kwargs), torch.nn.Flatten())  ## 1.

        data_dim = tuple([1] + list(image_out_size))
        self.model = ModelTackOn(
            model_chopped_pooled.to('cpu'),
            base_model_frozen.to('cpu'),
            data_dim=data_dim,
            pre_head_fc_sizes=pre_head_fc_sizes,
            post_head_fc_sizes=post_head_fc_sizes,
            nonlinearity=head_nonlinearity,
            kwargs_nonlinearity=head_nonlinearity_kwargs,
        )

        mnp = [name for name, param in self.model.named_parameters()]  ## 'model named parameters'
        mnp_blockNums = [name[name.find('.'):name.find('.')+8] for name in mnp]  ## pulls out the numbers just after the model name
        mnp_nums = [helpers.get_nums_from_string(name) for name in mnp_blockNums]  ## converts them to numbers
        block_to_freeze_nums = helpers.get_nums_from_string(block_to_unfreeze)  ## converts the input parameter specifying the block to freeze into a number for comparison

        m_baseName = mnp[0][:mnp[0].find('.')]

        for ii, (name, param) in enumerate(self.model.named_parameters()):
            if m_baseName in name:
                if mnp_nums[ii] < block_to_freeze_nums:
                    param.requires_grad = False
                elif mnp_nums[ii] >= block_to_freeze_nums:
                    param.requires_grad = True

        # self.model.forward = self.model.forward_latent if forward_version == 'forward_latent' else self.model.forward_head_pca
        self.model.forward = self.model.forward_latent if forward_version == 'forward_latent' else self.model.forward_head
    
    def save_onnx(
        self,
        check_load_onnx_valid: bool=False,
        revert_train: bool=True,
    ):
        """
        Uses ONNX to save the current model as a binary file.

        Args:
            check_load_onnx_valid (bool):
                Whether to check that the saved model is valid by loading onnx back in and
                comparing outputs
        """

        assert self.filepath_model_save is not None, 'You need to specify a filepath to save the model to.'

        batch_size = 1 # Arbitrary batch_size for saving the onnx model. Can be anything after load.

        import datetime
        try:
            import onnx
        except ImportError as e:
            raise ImportError(f'You need to (pip) install onnx to use this method. {e}')
        
        ## Make sure we have what we need
        assert self.model is not None, 'You need to fit the model first.'

        # Convert the model to ONNX format
        ## Prepare initial types

        # torch.onnx.export
        ## Prepare model
        self.model.eval()
        device_prev = self.model.device
        self.model.to('cpu')
        torch.onnx.export(
            self.model,
            (torch.ones(batch_size, 3, 224, 224),),
            self.filepath_model_save, # "onnx.pb",
            input_names=["x"],
            output_names=["latents"],
            dynamic_axes={
                # dict value: manually named axes
                "x": {0: "batch_size"},
                # list value: automatic names
                "latents": [0],
            }
        )
        self.model.to(device_prev)
        self.model.train()
        
        if check_load_onnx_valid:
            self.test(torch.ones((batch_size, 3, 224, 224), device='cpu'), revert_train=revert_train)

    def test(
        self,
        x,
        revert_train: bool=True,
        ):
        """
        Tests the model by loading the ONNX model and comparing outputs.

        Args:
            x (torch.Tensor):
                Input tensor to test the model with.
        """
    
        print('Checking ONNX model...')
        import onnxruntime as ort
        # Create example data
        # x = torch.ones((batch_size, 3, 224, 224))
        if hasattr(self.model, 'prep_contrast'):
            self.model.prep_contrast()
        self.model.eval()
        device_prev = self.model.device
        self.model.to('cpu')
        out_torch_original = self.model(x).cpu().detach().numpy()
        self.model.to(device_prev)
        
        model_loaded = self.load_onnx(self.filepath_model_save, inplace=False)
        model_loaded.eval()
        out_torch_loaded = model_loaded(x.cpu()).cpu().detach().numpy()

        # Check the Onnx output against PyTorch
        print(np.max(np.abs(out_torch_original - out_torch_loaded)))
        assert np.allclose(out_torch_original, out_torch_loaded, atol=1.e-5), "The outputs from the saved and loaded models are different."
        print('Saved ONNX model is valid.')

        if hasattr(self.model, 'prep_contrast'):
            self.model.prep_contrast()
        
        if revert_train:
            self.model.train()
            model_loaded.train()

    def load_onnx(
            self,
            filepath_model=None,
            inplace=True,
            ):
        """
        Loads the ONNX model from a file.

        Args:
            filepath_model (str):
                Path to the ONNX model file.
            inplace (bool):
                Whether to load the model as an attribute or return it.

        Returns:
            model (ModelTackOn):
                The loaded model.
        """
        
        assert filepath_model is not None, 'You need to specify a filepath from which to load the model.'

        try:
            import onnx
            import torch
            import onnx2torch
        except ImportError as e:
            raise ImportError(f'You need to (pip) install onnx and skl2onnx to use this method. {e}')
        
        # load ONNX model first
        if isinstance(filepath_model, str):
            model = onnx2torch.convert(filepath_model)
        else:
            raise ValueError(f'path_or_bytes must be either a string or bytes. This error should never be raised.')
        
        if inplace:
            self.model = model
        else:
            return model




import warnings


PACKAGES = [
    'GPUtil',
    'PIL',
    'cpuinfo',
    'cv2',
    'jupyter',
    'mat73',
    'matplotlib',
    'natsort',
    'numpy',
    'onnx',
    'onnxruntime',
    'optuna',
    'psutil',
    'pytest',
    'scipy',
    'seaborn',
    'skl2onnx',
    'sklearn',
    'sparse',
    'torch',
    'torchaudio',
    'torchvision',
    'tqdm',
    'xxhash',
    'yaml',
    'bokeh',
    'holoviews',
    'jupyter_bokeh',
    'umap',
    'hdbscan',
    'kymatio',
]

def test_internal_package_tests():
    """
    Test packages.
    RH 2023
    """
    from pathlib import Path

    import pytest

    ## List of packages to test
    packages = [
        'hdbscan',
        'umap',
        'optuna',
        'cv2',
        'skl2onnx',
        # 'holoviews',
        'bokeh',
        'sparse',
    ]

    ## Iterate over each package
    for pkg_s in packages:
        ## Try to import the package
        try:
            exec(f'import {pkg_s}')
            print(f'RH: Successfully imported {pkg_s}')
        except ImportError:
            warnings.warn(f'RH: Could not import {pkg_s}. Skipping tests.')
            continue

        else:
            try:
                ## Get a handle on the package
                pkg_h = eval(pkg_s)

                ## Get the path to the package
                path_pkg = str(Path(pkg_h.__file__).parent)

                ## Run the tests
                pytest.main([path_pkg, '-v'])
            except Exception as e:
                warnings.warn(f'RH: Could not run tests for {pkg_s}. Error: {e}')
                continue


def test_importing_packages():
    """
    Runs pytest on the core packages.
    """

    for pkg in PACKAGES:
        try:
            exec(f'import {pkg}')
        except ModuleNotFoundError:
            warnings.warn(f'RH Warning: {pkg} not found.')

def test_torch(
    device='cpu', 
    verbose=2
):
    """
    Test to see if torch can do operations on device.
    RH 2022

    Args:
        device (str):
            The device to use. Default is 'cuda'.
        verbose (int):
            If 0, do not print anything.
            If 1, print warnings.
            If 2, print all below and info.
    """
    import torch
    version = torch.__version__
        
    ## Test CPU computations
    arr = torch.rand(1000, 10, device=device)
    arr2 = torch.rand(10, 1000, device=device)
    arr3 = (arr @ arr2).mean().numpy()
    print(f'RH: Torch can do basic operations on CPU. Environment using PyTorch version: {version}. Result of operations: {arr3}') if verbose > 1 else None

def test_numpy():
    import numpy as np
    ## test numpy
    np.random.seed(0)
    arr1 = np.random.rand(1000, 10)
    arr2 = np.random.rand(10, 1000)
    arr3 = (arr1 @ arr2).mean()
    assert np.allclose(arr3, 2.5, rtol=0.1), 'RH Error: numpy test failed.'

def test_scipy_sparse():
    import numpy as np
    import scipy.sparse
    ## test scipy.sparse
    arr1 = scipy.sparse.rand(1000, 10, density=0.1).tocoo().tocsr()
    arr2 = scipy.sparse.rand(10, 1000, density=0.1).tocoo().tocsr()
    arr3 = (arr1 @ arr2).mean()
    assert np.allclose(arr3, 0.025, rtol=0.2), 'RH Error: scipy.sparse test failed.'



# Imports
import numpy as np
import torch
from PIL import Image
import matplotlib.pyplot as plt
import tqdm

import torch
import torch.cuda
from torch.autograd import Variable

from sklearn.metrics import accuracy_score
from sklearn.decomposition import PCA

from functools import partial
from typing import Optional, List, Tuple, Union, Dict, Any
from torch.nn import CrossEntropyLoss
from torch.optim import Adam

from . import model

def log_fn(log_str, log_file):
    """
    Write a string to a log file
    
    Args:
        log_str (str):
            String to be written to the log file
        log_file (str):
            Path to the log file
    """
    with open(log_file, 'a') as f:
        f.write(log_str + '\n')


class Simclr_Trainer():
    """
    Class to train a SimCLR model from scratch using the provided parameters.

    JZ 2023

    Args:
        dataloader (torch.utils.data.DataLoader):
            The dataloader to use for training.
        model_container (ModelContainer):
            The model container to use for training.
        training_stop_revert_atNan (bool):
            Whether to revert to the previous model if the loss becomes NaN and stop training.
        n_epochs (int):
            The number of epochs to train for.
        device_train (str):
            The device to train on.
        inner_batch_size (int):
            The batch size to use for training.
        learning_rate (float):
            The learning rate to use for training.
        penalty_orthogonality (float):
            The penalty to apply to the orthogonality of the latent space.
        weight_decay (float):
            The weight decay to use for training.
        gamma (float):
            The gamma to use for training.
        temperature (float):
            The temperature to use for training.
        l2_alpha (float):
            The alpha to use for L2 regularization.
        path_saveLog (Optional[str]):
            The path to which to save the training log.
        path_saveLoss (Optional[str]):
            The path to which to save the losses.
    """


    def __init__(
            self,
            dataloader: torch.utils.data.DataLoader,
            model_container: model.Simclr_Model,
            
            training_stop_revert_atNan: bool = True,
                    
            n_epochs: int = 9999999,
            device_train: str = 'cuda:0',
            inner_batch_size: int = 256,
            learning_rate: float = 0.01,
            penalty_orthogonality: float = 1.00,
            weight_decay: float = 0.1,
            gamma: float = 1.0000,
            temperature: float = 0.03,
            l2_alpha: float = 0.0000,

            path_saveLog: Optional[str] = None,
            path_saveLoss: Optional[str] = None,
        ):

        self.dataloader = dataloader
        self.training_stop_revert_atNan = training_stop_revert_atNan
        self.model_container = model_container
        self.n_epochs = n_epochs
        self.device_train = device_train
        self.inner_batch_size = inner_batch_size
        self.learning_rate = learning_rate
        self.penalty_orthogonality = penalty_orthogonality
        self.weight_decay = weight_decay
        self.gamma = gamma
        self.temperature = temperature
        self.l2_alpha = l2_alpha
        self.path_saveLog = path_saveLog
        self.path_saveLoss = path_saveLoss

    def train(
            self
            ):
        """
        Trains the model using the saved attributes.
        """
        self.model_container.model.train();
        self.model_container.model.to(self.device_train)
        self.model_container.model.prep_contrast()

        criteria = [CrossEntropyLoss()]
        criteria = [criterion.to(self.device_train) for criterion in criteria]
        optimizer = Adam(
            self.model_container.model.parameters(), 
            lr=self.learning_rate,
        )
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer,
                                                        gamma=self.gamma,
                                                        )
        
        self.dataloader
        self.model_container
        self.n_epochs
        self.device_train
        self.inner_batch_size
        self.learning_rate
        self.penalty_orthogonality
        self.weight_decay
        self.gamma
        self.temperature
        self.l2_alpha
        self.path_saveLog

        log_function = partial(log_fn, log_file=self.path_saveLog) if self.path_saveLog is not None else lambda x: None

        losses_train, losses_val = [], [np.nan]
        for epoch in tqdm.tqdm(range(self.n_epochs)):
            print(f'epoch: {epoch}')
            losses_train = epoch_step(
                self.dataloader, 
                self.model_container.model, 
                optimizer, 
                criteria,
                scheduler=scheduler,
                temperature=self.temperature,
                penalty_orthogonality=self.penalty_orthogonality,
                loss_rolling_train=losses_train, 
                loss_rolling_val=losses_val,
                device=self.device_train, 
                inner_batch_size=self.inner_batch_size,
                verbose=2,
                verbose_update_period=1,
                log_function=log_function,
            )
            
            ## save loss information
            if self.path_saveLoss is not None:
                np.save(self.path_saveLoss, losses_train)
            
            ## if loss becomes NaNs, don't save the network and stop training
            if torch.isnan(torch.as_tensor(losses_train[-1])) and self.training_stop_revert_atNan:
                break

            ## save model
            self.model_container.save_onnx(check_load_onnx_valid=True)

def train_step_simCLR( 
    X_train_batch, 
    y_train_batch, 
    model, 
    optimizer,
    criterion, 
    scheduler, 
    temperature, 
    sample_weights,
    penalty_orthogonality=0,
    inner_batch_size=None,
    ):
    """
    Performs a single training step.
    RH 2021 / JZ 2021

    Args:
        X_train_batch (torch.Tensor):
            Batch of training data.
            Shape: 
             (batch_size, n_transforms, n_channels, height, width)
        y_train_batch (torch.Tensor):
            Batch of training labels
            NOT USED FOR NOW
        model (torch.nn.Module):
            Model to train
        optimizer (torch.optim.Optimizer):
            Optimizer to use
        criterion (torch.nn.Module):
            Loss function to use
        scheduler (torch.optim.lr_scheduler.LambdaLR):
            Learning rate scheduler
        temperature (float):
            Temperature term for the softmax
        sample_weights (torch.Tensor):
            Sample weights for the batch
        penalty_orthogonality (float):
            Penalty for the orthogonality of the weights
        inner_batch_size (Optional[int]):
            Batch size for the inner loop. If None, the whole batch is used.
    
    Returns:
        loss (float):
            Loss of the current batch
        pos_over_neg (float):
            Ratio of logits of positive to negative samples
    """

    double_sample_weights = torch.tile(sample_weights.reshape(-1), (2,))
    contrastive_matrix_sample_weights = torch.cat((torch.ones(1, device=X_train_batch.device), double_sample_weights), dim=0)
    
    optimizer.zero_grad()

    if inner_batch_size is None:
        features = model.forward_latent(X_train_batch)
    else:
        features = torch.cat([model.forward_latent(sub_batch) for sub_batch in make_batches(X_train_batch, batch_size=inner_batch_size)], dim=0)
    
    torch.cuda.empty_cache()

    logits, labels = richs_contrastive_matrix(features, batch_size=X_train_batch.shape[0]/2, n_views=2, temperature=temperature, DEVICE=X_train_batch.device) #### FOR RICH - THIS IS THE LINE IN QUESTION. I THINK "/2" NEEDS TO BE REMOVED FROM "X_train_batch.shape[0]/2"
    pos_over_neg = (torch.mean(logits[:,0]) / torch.mean(logits[:,1:])).item()

    loss_unreduced_train = torch.nn.functional.cross_entropy(logits, labels, weight=contrastive_matrix_sample_weights, reduction='none')
    loss_train = (loss_unreduced_train.float() @ double_sample_weights.float()) / double_sample_weights.float().sum()
    loss_orthogonality = off_diagonal(torch.corrcoef(features.T)).pow_(2).sum().div(features.shape[1])
    loss = loss_train + penalty_orthogonality * loss_orthogonality


    loss.backward()
    optimizer.step()
    scheduler.step()

    return loss.item(), pos_over_neg

def off_diagonal(x):
    """
    Returns the off-diagonal elements of a matrix as a vector.
    RH 2022

    Args:
        x (np.ndarray or torch tensor):
            square matrix to extract off-diagonal elements from.

    Returns:
        output (np.ndarray or torch tensor):
            off-diagonal elements of x.
    """
    n, m = x.shape
    assert n == m
    return x.reshape(-1)[:-1].reshape(n - 1, n + 1)[:, 1:].reshape(-1)

def L2_reg(model):
    """
    Returns the L2 regularization penalty for a model.
    RH 2021 / JZ 2021

    Args:
        model (torch.nn.Module):
            Model to calculate the penalty for.

    Returns:
        penalty (float):
            L2 regularization penalty.
    """
    penalized_params = get_trainable_parameters(model)
    penalty = 0
    for ii, param in enumerate(penalized_params):
        penalty += torch.sum(param**2)
    return penalty

def epoch_step( dataloader, 
                model, 
                optimizer, 
                criterion, 
                scheduler=None, 
                temperature=0.5,
                penalty_orthogonality=0,
                loss_rolling_train=[], 
                loss_rolling_val=[],
                device='cpu', 
                inner_batch_size=None,
                verbose=False,
                verbose_update_period=100,
                log_function=print
                ):
    """
    Performs an epoch step.
    RH 2021 / JZ 2021

    Args:
        dataloader (torch.utils.data.DataLoader):
            Dataloader for the current epoch.
            Output for X_batch should be shape:
             (batch_size, n_transforms, n_channels, height, width)
        model (torch.nn.Module):
            Model to train
        optimizer (torch.optim.Optimizer):
            Optimizer to use
        criterion (torch.nn.Module):
            Loss function to use
        scheduler (torch.optim.lr_scheduler.LambdaLR):
            Learning rate scheduler
        temperature (float):
            Temperature term for the softmax
        penalty_orthogonality (float):
            Penalty for the orthogonality of the weights
        loss_rolling_train (list):
            List of losses for the current epoch
        loss_rolling_val (list):
            List of losses for the validation set
        device (str):
            Device to run the loss on
        inner_batch_size (Optional[int]):
            Batch size for the inner loop. If None, the whole batch is used.
        verbose (bool):
            Whether to print out the loss
        verbose_update_period (int):
            How often to print out the loss
        log_function (function):
            Function to use for printing out the loss

    Returns:
        loss_rolling_train (list):
            List of losses (passed through and appended)
    """

    def print_info(batch, n_batches, loss_train, loss_val, pos_over_neg, learning_rate, precis=5):
        log_function(f'Iter: {batch}/{n_batches}, loss_train: {loss_train:.{precis}}, loss_val: {loss_val:.{precis}}, pos_over_neg: {pos_over_neg} lr: {learning_rate:.{precis}}')

    for i_batch, (X_batch, y_batch, idx_batch, sample_weights) in enumerate(dataloader):
        X_batch = torch.cat(X_batch, dim=0)
        X_batch = X_batch.to(device)
        y_batch = y_batch.to(device)
        
        # Get batch weights
        loss, pos_over_neg = train_step_simCLR(
            X_batch, 
            y_batch, 
            model, 
            optimizer, 
            criterion, 
            scheduler, 
            temperature, 
            sample_weights=torch.as_tensor(sample_weights, device=device),
            penalty_orthogonality=penalty_orthogonality,
            inner_batch_size=inner_batch_size,
            ) # Needs to take in weights
        loss_rolling_train.append(loss)
        # if False and do_validation:
        #     loss = validation_Object.get_predictions()
        #     loss_rolling_val.append(loss)
        if verbose>0:
            if i_batch%verbose_update_period == 0:
                print_info( batch=i_batch,
                            n_batches=len( dataloader),
                            loss_train=loss_rolling_train[-1],
                            loss_val=loss_rolling_val[-1],
                            pos_over_neg=pos_over_neg,
                            learning_rate=scheduler.get_last_lr()[0],
                            precis=5)
    return loss_rolling_train

# # from https://github.com/sthalles/SimCLR/blob/1848fc934ad844ae630e6c452300433fe99acfd9/simclr.py
def info_nce_loss(features, batch_size, n_views=2, temperature=0.5, DEVICE='cpu'):
    """
    'Noise-Contrastice Estimation' loss. Loss used in SimCLR.
    InfoNCE loss. Aka: NTXentLoss or generalized NPairsLoss.
    
    logits and labels should be run through 
     CrossEntropyLoss to complete the loss.
    CURRENTLY ONLY WORKS WITH n_views=2 (I think this can be
     extended to larger numbers by simply reshaping logits)

    Code mostly copied from: https://github.com/sthalles/SimCLR/blob/1848fc934ad844ae630e6c452300433fe99acfd9/simclr.py
    RH 2021 / JZ 2021

    demo for learning/following shapes:
        features = torch.rand(8, 100)
        labels = torch.cat([torch.arange(4) for i in range(2)], dim=0)

        labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
        print(labels)
        features = torch.nn.functional.normalize(features, dim=1)
        similarity_matrix = torch.matmul(features, features.T)
        print(similarity_matrix)
        mask = torch.eye(labels.shape[0], dtype=torch.bool)
        labels = labels[~mask].view(labels.shape[0], -1)
        print(labels)
        similarity_matrix = similarity_matrix[~mask].view(similarity_matrix.shape[0], -1)
        print(similarity_matrix)
        positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)
        print(positives)
        negatives = similarity_matrix[~labels.bool()].view(similarity_matrix.shape[0], -1)
        print(negatives)
        logits = torch.cat([positives, negatives], dim=1)
        print(logits)
        labels = torch.zeros(logits.shape[0], dtype=torch.long)
        print(labels)

    Args:
        features (torch.Tensor): 
            Outputs of the model
            Shape: (batch_size * n_views, n_channels, height, width)
        batch_size (int):
            Number of samples in the batch
        n_views (int):
            Number of views in the batch
            MUST BE 2 (larger values not supported yet)
        temperature (float):
            Temperature term for the softmax
        DEVICE (str):
            Device to run the loss on
    
    Returns:
        logits (torch.Tensor):
            Class prediction logits
            Shape: (batch_size * n_views, ((batch_size-1) * n_views)+1)
    """

    # make (double) diagonal matrix
    labels = torch.cat([torch.arange(batch_size) for i in range(n_views)], dim=0)
    labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
    labels = labels.to(DEVICE)

    # normalize to unit hypersphere
    features = torch.nn.functional.normalize(features, dim=1)

    # compute (double) covariance matrix
    similarity_matrix = torch.matmul(features, features.T)
    # assert similarity_matrix.shape == (
    #     self.args.n_views * self.args.batch_size, self.args.n_views * self.args.batch_size)
    # assert similarity_matrix.shape == labels.shape

    # discard the main diagonal from both: labels and similarities matrix
    mask = torch.eye(labels.shape[0], dtype=torch.bool).to(DEVICE)
    labels = labels[~mask].view(labels.shape[0], -1)
    similarity_matrix = similarity_matrix[~mask].view(similarity_matrix.shape[0], -1)
    # assert similarity_matrix.shape == labels.shape

    # select and combine multiple positives
    positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)

    # select only the negatives the negatives
    negatives = similarity_matrix[~labels.bool()].view(similarity_matrix.shape[0], -1)

    logits = torch.cat([positives, negatives], dim=1) # logits column 1 is positives, the rest of the columns are negatives
    labels = torch.zeros(logits.shape[0], dtype=torch.long).to(DEVICE) # all labels are 0 because first column in logits is positives

    logits = logits / temperature
    return logits, labels


# # from https://github.com/sthalles/SimCLR/blob/1848fc934ad844ae630e6c452300433fe99acfd9/simclr.py
def richs_contrastive_matrix(features, batch_size, n_views=2, temperature=0.5, DEVICE='cpu'):
    """
    Modified 'Noise-Contrastice Estimation' loss. 
    Almost identical to the method used in SimCLR.
    Should be techincally identical to InfoNCE, 
     but the output logits matrix is different.
    The output logits first column is the positives,
     and the rest of the columns are the cosine 
     similarities of the negatives (positives )
    InfoNCE loss. Aka: NTXentLoss or generalized NPairsLoss.
    
    logits and labels should be run through 
     CrossEntropyLoss to complete the loss.
    CURRENTLY ONLY WORKS WITH n_views=2 (I think this can be
     extended to larger numbers by simply reshaping logits)

    Code mostly copied from: https://github.com/sthalles/SimCLR/blob/1848fc934ad844ae630e6c452300433fe99acfd9/simclr.py
    RH 2021

    demo for learning/following shapes:
        features = torch.rand(8, 100)
        eye_prep = torch.cat([torch.arange(4) for i in range(2)], dim=0)
        print(eye_prep)
        multi_eye = (eye_prep.unsqueeze(0) == eye_prep.unsqueeze(1))
        print(multi_eye)
        features = torch.nn.functional.normalize(features, dim=1)
        similarity_matrix = torch.matmul(features, features.T)
        print(similarity_matrix)
        single_eye = torch.eye(multi_eye.shape[0], dtype=torch.bool)
        multi_cross_eye = multi_eye * ~single_eye
        print(multi_cross_eye)
        positives = similarity_matrix[multi_cross_eye.bool()].view(multi_cross_eye.shape[0], -1)
        print(positives)
        logits = torch.cat([positives, similarity_matrix*(~multi_eye)], dim=1)
        print(logits)
        labels = torch.zeros(logits.shape[0], dtype=torch.long)
        print(labels)

    Args:
        features (torch.Tensor): 
            Outputs of the model
            Shape: (batch_size * n_views, n_channels, height, width)
        batch_size (int):
            Number of samples in the batch
        n_views (int):
            Number of views in the batch
            MUST BE 2 (larger values not supported yet)
        temperature (float):
            Temperature term for the softmax
        DEVICE (str):
            Device to run the loss on
    
    Returns:
        logits (torch.Tensor):
            Class prediction logits
            Shape: (batch_size * n_views, 1 + (batch_size * n_views))
    """


    eye_prep = torch.cat([torch.arange(batch_size) for i in range(n_views)], dim=0).to(DEVICE)
    multi_eye = (eye_prep.unsqueeze(0) == eye_prep.unsqueeze(1))
    features = torch.nn.functional.normalize(features, dim=1)
    similarity_matrix = torch.matmul(features, features.T)
    single_eye = torch.eye(multi_eye.shape[0], dtype=torch.bool).to(DEVICE)
    multi_cross_eye = multi_eye * ~single_eye
    positives = similarity_matrix[multi_cross_eye].view(multi_cross_eye.shape[0], -1)
    logits = torch.cat([positives, similarity_matrix*(~multi_eye)], dim=1)
    labels = torch.zeros(logits.shape[0], dtype=torch.long).to(DEVICE)

    logits = logits / temperature
    return logits, labels

def make_batches(iterable, batch_size=None, num_batches=5, min_batch_size=0, return_idx=False):
    """
    Make batches of data or any other iterable.
    RH 2021

    Args:
        iterable (iterable):
            iterable to be batched
        batch_size (int):
            size of each batch
            if None, then batch_size based on num_batches
        num_batches (int):
            number of batches to make
        min_batch_size (int):
            minimum size of each batch
        return_idx (bool):
            whether to return the indices of the batches.
            output will be [start, end] idx
    
    Returns:
        output (iterable):
            batches of iterable
    """
    l = len(iterable)
    
    if batch_size is None:
        batch_size = np.int64(np.ceil(l / num_batches))
    
    for start in range(0, l, batch_size):
        end = min(start + batch_size, l)
        if (end-start) < min_batch_size:
            break
        else:
            if return_idx:
                yield iterable[start:end], [start, end]
            else:
                yield iterable[start:end]

def get_trainable_parameters(model):
    """
    Get the trainable parameters of a model.

    Args:
        model (torch.nn.Module):
            model to get trainable parameters from

    Returns:
        params_trainable (list):
            list of trainable parameters
    """
    params_trainable = []
    for param in list(model.parameters()):
        if param.requires_grad:
            params_trainable.append(param)
    return params_trainable

class Simclr_PCA_Trainer():
    def __init__(
            self,
            dataloader,
            model_container,

            center: bool = True,
            scale: bool = False,
            path_saveLog: Optional[str] = None,

            # use_iterated_learning: bool = False,
            ):
        """
        Training module to train a SimCLR model from scratch using the provided parameters.

        JZ 2023

        Args:
            dataloader (torch.utils.data.DataLoader):
                The dataloader to use for training.
            model_container (ModelContainer):
                The model container to use for training.
            center (bool):
                Whether to center the data.
            scale (bool):
                Whether to scale the data.
            path_saveLog (str):
                The path to which to save the training log.
        """

        self.dataloader = dataloader
        self.model_container = model_container
        self.center = center
        self.scale = scale
        self.path_saveLog = path_saveLog

    def train(self, check_pca_layer_valid: bool=True):
        """
        Trains the pca layer of the model using the input data x and saves
        the updated model to self.model_container.filepath_model_save

        Args:
            check_pca_layer_valid (bool):
                Whether to check that the pca layer is valid after training.
        """

        # x = torch.cat([torch.cat(data_subset[0], dim=0) for data_subset in self.dataloader], axis=0)
        # output_head = self.model_container.model(x)
        output_head = torch.cat([self.model_container.model(torch.cat(data_subset[0], dim=0)) for data_subset in self.dataloader], dim=0)
        pca_size = output_head.shape[1]

        output_head_scaler = output_head.std(dim=0) if self.scale else torch.ones_like(output_head.std(dim=0))
        output_head_centerer = output_head.mean(dim=0)/output_head_scaler if self.center else torch.zeros_like(output_head.mean(dim=0)/output_head_scaler)

        pca_layer = torch.nn.Sequential(
            torch.nn.Linear(pca_size, pca_size),
            torch.nn.Linear(pca_size, pca_size, bias=False)
        )
        pca_layer[0].weight = torch.nn.Parameter(torch.diag(1/output_head_scaler))
        pca_layer[0].bias = torch.nn.Parameter(-output_head_centerer)

        output_head_centered = (output_head - output_head_centerer) / output_head_scaler
        
        if check_pca_layer_valid:
            assert torch.allclose(
                output_head_centered,
                pca_layer[0](output_head),
                atol=torch.tensor(1e-4)
            ), 'zscore layer not working'

        np_output_head_centered = output_head_centered.detach().cpu().numpy()

        pca_sklearn = PCA()
        pca_sklearn.fit(np_output_head_centered)
        pca_sklearn.components_ = pca_sklearn.components_.astype(np.float32)
        pca_sklearn.mean_ = np.zeros(pca_size, dtype=np.float32)

        pca_layer[1].weight = torch.nn.Parameter(torch.tensor(pca_sklearn.components_, dtype=torch.float32))
        # pca_layer[1].bias = torch.nn.Parameter(torch.tensor(np.zeros(pca_size,),dtype=torch.float32))

        if check_pca_layer_valid:
            pca_output = pca_layer(output_head)
            pca_sklearn_tensor = torch.tensor(pca_sklearn.transform(np_output_head_centered))
            print(np.max(np.abs(pca_output.detach().numpy() - pca_sklearn_tensor.detach().numpy())))
            assert torch.allclose(pca_output,
                            pca_sklearn_tensor,
                            atol=1e-5
                            ), 'pca layer not working'
        
        print('save')

        ## save model
        class ModelPCATackOn(torch.nn.Module):
            def __init__(self, model_pre, pca_layer):
                super().__init__()
                self.model = torch.nn.Sequential(
                    model_pre,
                    pca_layer
                )
            @property
            def device(self):
                return next(self.model.parameters()).device
            def forward(self, x):
                return self.model(x)
        self.model_container.model = ModelPCATackOn(self.model_container.model, pca_layer)

        self.model_container.save_onnx(check_load_onnx_valid=True, revert_train=False)

        print('pca layer trained and saved to model_container')



import gc
from typing import Any, Dict, Tuple

import torch
import numpy as np
from tqdm import tqdm

from .. import helpers, util

class SWT(util.ROICaT_Module):
    """
    Performs scattering wavelet transform using the kymatio library.
    RH 2022

    Args:
        kwargs_Scattering2D (Dict[str, Any]):
            The keyword arguments to pass to the Scattering2D class.
            (Default is ``{'J': 2, 'L': 8}``)
        image_shape (Tuple[int, int]):
            The shape of the images to be transformed. 
            (Default is ``(36,36)``)
        device (str):
            The device to use for the transformation. 
            (Default is ``'cpu'``)
        verbose (bool):
            If ``True``, print statements will be outputted. 
            (Default is ``True``)
    
    Example:
        .. highlight:: python
        .. code-block:: python

            swt = SWT(kwargs_Scattering2D={'J': 2, 'L': 8}, image_shape=(36,36), device='cpu', verbose=True)
            transformed_images = swt.transform(ROI_images, batch_size=100)
    """
    def __init__(
        self, 
        kwargs_Scattering2D: Dict[str, Any] = {'J': 2, 'L': 8}, 
        image_shape: Tuple[int, int] = (36,36), 
        device: str = 'cpu',
        verbose: bool = True,
    ):
        """
        Initializes the SWT with the given settings.
        """
        ## Imports
        super().__init__()

        ## Store parameter (but not data) args as attributes
        self.params['__init__'] = self._locals_to_params(
            locals_dict=locals(),
            keys=[
                'kwargs_Scattering2D',
                'image_shape',
                'device',
                'verbose',
            ],
        )

        from kymatio.torch import Scattering2D

        self._verbose = verbose
        self._device = device
        self.swt = Scattering2D(shape=image_shape, **kwargs_Scattering2D)
        self.swt = util.Model_SWT(self.swt)
        self.swt.to(device)
        print('SWT initialized') if self._verbose else None

    def transform(self, ROI_images: np.ndarray, batch_size: int = 100) -> np.ndarray:
        """
        Transforms the ROI images.

        Args:
            ROI_images (np.ndarray):
                The ROI images to transform. 
                One should probably concatenate ROI images across sessions for passing through here. 
                *(n_ROIs, height, width)*
            batch_size (int):
                The batch size to use for the transformation. 
                (Default is *100*)

        Returns:
            (np.ndarray):
                latents (np.ndarray):
                    The transformed ROI images. *(n_ROIs, latent_size)*
        """
        ## Store parameter (but not data) args as attributes
        self.params['transform'] = self._locals_to_params(
            locals_dict=locals(),
            keys=['batch_size',],)

        print('Starting: SWT transform on ROIs') if self._verbose else None
        def helper_swt(ims_batch):
            sfs = torch.as_tensor(np.ascontiguousarray(ims_batch[None,...]), device=self._device, dtype=torch.float32)
            out = self.swt(sfs[None,...]).squeeze().cpu()
            if out.ndim == 3:  ## if there is only one ROI in the batch, append a dimension to the front
                out = out[None,...]
            return out
        self.latents = torch.cat([helper_swt(ims_batch) for ims_batch in tqdm(helpers.make_batches(ROI_images, batch_size=batch_size), total=ROI_images.shape[0] / batch_size, mininterval=5)], dim=0)
        self.latents = self.latents.reshape(self.latents.shape[0], -1)
        print('Completed: SWT transform on ROIs') if self._verbose else None

        gc.collect()
        torch.cuda.empty_cache()
        gc.collect()
        torch.cuda.empty_cache()

        return self.latents
